/* s29gl512n_16x2.c
 * Device interface for Spansion S29GL512N flash device configured for
 * x16 mode with 2 devices in parallel.
 * The device contains 512 128Kbyte sectors.
 *
 * TODO: This whole file is a gross hack.  What we really should be
 * doing is invoking the 16x1 drivers, after first figuring out which
 * of the two chips (high word or low word) a specified write access
 * is referring to. Erases should always affect both chips, as though
 * each sector was 256K in size.
 *
 * TODO: fdev->width is the width of our flash data bus, which doesn't
 * necessarily equal the width of the flash chip's data bus (e.g. two
 * chips in parallel to double the width of the bus). Ideally, we'd
 * program in the flash chip's data bus width, so that we write only
 * one driver and just re-use it across both chips.  But then we can't
 * as easily take advantage of being able to hit two flashes in
 * parallel if the hardware works that way.  Furthermore, we want to
 * be able to take advantage of each chip's write buffer if the chip
 * supports that feature.  Hmmm, this abstraction (and
 * implementation!) still needs work...
 */
#include "config.h"

#if INCLUDE_FLASH

#define WRITE_BUFFER_SIZE	32
#define WRITE_BUFFER_ADDR_MASK		(~(WRITE_BUFFER_SIZE-1))

#include "stddefs.h"
#include "genlib.h"	
#include "cpu.h"
#include "cpuio.h"
#include "flash.h"			/* Part of monitor common code */
#include "s29gl512n_16x1.h"

#define SECTOR_TOTAL		512
#define SECTOR_SIZE		0x40000
#define SPANSION_29GL512N_ID	0x00012223

#define D23	(1 << 23)
#define D7	(1 << 7)
#define D22	(1 << 22)
#define D6	(1 << 6)
#define D21	(1 << 21)
#define D5	(1 << 5)
#define D19	(1 << 19)
#define D3	(1 << 3)

/* TODO: these macros work only when ftype is a long; fix that */
#define ftype				volatile unsigned long
#define Is_ff(add)			(*((ftype *)(add)) == 0xffffffffUL)
#define Is_Not_ff(add)			(!(Is_ff(add)))
#define Is_Equal(p1,p2)			(*((ftype *)(p1)) == *((ftype *)(p2)))
#define Is_Equal_d7(p1,p2)		((*((ftype *)(p1)) & (D7 | D23)) == (*((ftype *)(p2)) & (D7 | D23)))
#define Is_Not_Equal(p1,p2)		(!(Is_Equal((p1),(p2))))
#define Is_Not_Equal_d7(p1,p2)		(!(Is_Equal_d7((p1),(p2))))

#define Fwrite(to,frm)		(*((ftype *)(to)) = *((ftype *)(frm)))
#define D5_Timeout(add)		((*((ftype *)(add)) & (D5 | D21)) == (D5 | D21))

/* TODO: (fdev->width / 2) works for all widths, but the commands we
 * write are all 32 bits in size in the following macros; we should
 * instead truncate the command to fdev->width bytes so that the upper
 * bytes get removed when in 16/8-bit mode
 */

#define SECTOR_ERASE(add)		{ \
	*(ftype *)((fdev->base + (0x555 << (fdev->width / 2)))) = 0xaa00aaUL; \
	*(ftype *)((fdev->base + (0x2aa << (fdev->width / 2)))) = 0x550055UL; \
	*(ftype *)((fdev->base + (0x555 << (fdev->width / 2)))) = 0x800080UL; \
	*(ftype *)((fdev->base + (0x555 << (fdev->width / 2)))) = 0xaa00aaUL; \
	*(ftype *)((fdev->base + (0x2aa << (fdev->width / 2)))) = 0x550055UL; \
	*(ftype *)(add) = 0x300030; }
	
#define FLASH_WRITE(dest, src)		{ \
	*(ftype *)((fdev->base + (0x555 << (fdev->width / 2)))) = 0xaa00aaUL; \
	*(ftype *)((fdev->base + (0x2aa << (fdev->width / 2)))) = 0x550055UL; \
	*(ftype *)((fdev->base + (0x555 << (fdev->width / 2)))) = 0xa000a0UL; \
	Fwrite((ftype *)(dest),src); }

#define FLASH_INIT_BUFFER(dest, bytes)	{ \
	*(ftype *)((fdev->base + (0x555 << (fdev->width / 2)))) = 0xaa00aaUL; \
	*(ftype *)((fdev->base + (0x2aa << (fdev->width / 2)))) = 0x550055UL; \
	*(ftype *)(dest) = 0x250025UL; \
	*(ftype *)(dest) = (((bytes) / (fdev->width / 2)) - 1); }
	
#define FLASH_WRITE_BUFFER(dest)	\
	*(ftype *)((fdev->base + (0x555 << (fdev->width / 2)))) = 0x290029UL;

#define AUTO_SELECT()			{ \
	*(ftype *)((fdev->base + (0x555 << (fdev->width / 2)))) = 0xaa00aa; \
	*(ftype *)((fdev->base + (0x2aa << (fdev->width / 2)))) = 0x550055; \
	*(ftype *)((fdev->base + (0x555 << (fdev->width / 2)))) = 0x900090; }

#define READ_RESET() 			{ \
	ftype val;				\
	*(ftype *)(fdev->base) = 0xf000f0;	\
	val = *(ftype *)(fdev->base);		}

/* TODO: what if one flash finishes earlier than the other? */
#define WHILE_D6_TOGGLES(a) \
	while ((*(ftype *)(a) & (D6 | D22)) != (*(ftype *)(a) & (D6 | D22)))


/* S29gl512n_16x2_erase():
 * Based on the 'snum' value, erase the appropriate sector(s).
 * Return 0 if success, else -1.
 */
int
S29gl512n_16x2_erase(struct flashinfo *fdev,int snum)
{
	ulong add;
	int ret;

	ret = 0;
	add = (ulong)(fdev->sectors[snum].begin);


	SECTOR_ERASE(add);

	/* Wait for sector erase to complete or timeout..
	 * DQ7 polling: wait for D7 to be 1.
	 * DQ6 toggling: wait for D6 to not toggle.
	 * DQ5 timeout: if DQ7 is 0, and DQ5 = 1, timeout.
	 */
	while(1) {
		WATCHDOG_MACRO;
		if (Is_ff(add) && Is_ff(add))
			break;

		if (D5_Timeout(add)) {
			if (Is_Not_ff(add))
				ret = -1;
			break;
		}
	}

	READ_RESET();

	return ret;
}

/* EndS29gl512n_16x2_erase():
 * Function place holder to determine the end of the above function.
 */
void
EndS29gl512n_16x2_erase(void)
{
}

/* what I'm really after is a simple modulo operation, but I can't use
 * that operator or gcc will call a libgcc helper function; this cheat
 * will work only because my modulos are always powers of two */
#define PTR_ALIGN(p, a) (((unsigned long)(p)) & ((a) - 1))

/* S29gl512n_16x2_write():
 * Copy specified number of bytes from source to destination.  The destination
 * address is assumed to be flash space.
 *
 * TODO: src and dest are really unions of the different pointer
 * types, and handling them that way would eliminate most of the casts
 * in this code
 */
int
S29gl512n_16x2_write(struct flashinfo *fdev,
		     uchar *dest, uchar *src, long bytecnt)
{
	int ret = 0;
	int wbyte;

	/* TODO: these should really be a union! */
	ftype val;
	uchar * const buf = (uchar *)&val;

	/* how much we moved dest to align it */
	int doft = PTR_ALIGN(dest, fdev->width);

	/* destination address must always be aligned to fdev->width,
	 * because that's how the hardware works */
	dest -= doft;

	do {
		/* the flash hardware forces us to write in
		 * address-aligned units of fdev->width; this loop
		 * gathers up an aligned unit of data from src, and
		 * writes it to dest; if such a unit isn't available,
		 * we fetch enough bytes from dest to pad out the
		 * difference */

		val = *(ftype *)dest;

		if (doft && doft != fdev->width)
		{
			/* we had to align dest above, so read bytes
			 * into dest buffer to account for that (this
			 * should happen only on the first iteration
			 * when dest was originally misaligned) */
			while (doft < fdev->width && bytecnt--)
				buf[doft++] = *src++;
		}
		else if (PTR_ALIGN(src, fdev->width) || bytecnt < fdev->width)
		{
			/* src is misaligned, or we have too few bytes
			 * to do an aligned-width transfer */
			for (wbyte = 0;
			     wbyte < fdev->width && bytecnt;
			     bytecnt--, wbyte++)
				buf[wbyte] = *src++;
		}			
		else
		{
			/* src and dest are properly aligned, and we
			 * have enough bytes to do a full-width
			 * transfer --- yay! */
			val = *(ftype *)src;
			src += fdev->width;
			bytecnt -= fdev->width;
		}

		/* TODO: how to take advantage of write buffering?
		* One idea is to have the fdev structure indicate the
		* size of the buffer, and have the calling function
		* pre-align and organize everything into buffer-sized
		* chunks... */
		
		FLASH_WRITE(dest, buf);

		while (1)
		{
			WATCHDOG_MACRO;

			/* TODO: this is a clever way of checking for
			 * completion, but it assumes that the flash
			 * drops immediately back into read-array mode
			 * at the end of the programming cycle; not
			 * all chips do that (the s29's do, tho) */
			if (*(ftype *)dest == val
			    && *(ftype *)dest == val)
				break;

			/* TODO: I'm not quite sure the logic here is
			 * right; what if the flash chips don't finish
			 * at the same time, or one times out but the
			 * other doesn't? */
			if (D5_Timeout(dest)
			    && *(ftype*)dest != val) {
				ret = -1;
				goto done;
			}
		}

		dest += fdev->width;
	} while (bytecnt);

done:
	READ_RESET();

	return(ret);
}

/* EndS29gl512n_16x2_write():
 * Function place holder to determine the end of the above function.
 */
void
EndS29gl512n_16x2_write(void)
{}

/* S29gl512n_16x2_ewrite():
 * Erase all sectors that are part of the address space to be written,
 * then write the data to that address space.
 * This function is optimized for speed by using the write buffer and
 * erase delay timer.  It assumes the incoming destination address is
 * aligned on a 32-byte boundary and that the bytecount will be even.
 */
int
S29gl512n_16x2_ewrite(struct flashinfo *fdev,
		      uchar *dest, uchar *src, long bytecnt)
{
	int wsec, wbyte;
	struct sectorinfo *s;

	/* TODO: these should really be a union! */
	ftype val;
	uchar * const buf = (uchar *)&val;

	/* how much we moved dest to align it */
	int doft = PTR_ALIGN(dest, fdev->width);

	/* erase each sector that encloses the region of interest */
	for (wsec = 0, s = fdev->sectors;
	     wsec < fdev->sectorcnt;
	     wsec++, s++)
	{
		struct sectorinfo *s = &(fdev->sectors[wsec]);

		if ((dest > s->end)
		    || ((dest + bytecnt - 1) < s->begin))
			continue;
		
		SECTOR_ERASE(s->begin);
		WHILE_D6_TOGGLES(s->begin);
	}

	READ_RESET();

#if 0
	/* TODO: this code uses the write buffer; try to re-use it */

	while(bytecnt > 0) {
		int tot, tot1;
		uchar *sa;

		sa = dest;
		*(ftype *)((fdev->base+(0x555<<1))) = 0x00aa;
		*(ftype *)((fdev->base+(0x2aa<<1))) = 0x0055;

		*(ftype *)(sa) = 0x0025;
		if (bytecnt >= 32)
			tot = tot1 = 32;
		else
			tot = tot1 = bytecnt;
		*(ftype *)(dest) = (tot/2 - 1);
		while(tot > 0) {
			Fwrite((ftype *)(dest),src); 
			dest+=2; 
			src+=2;
			tot -= 2;
		}
  	    *(ftype *)(sa) = 0x0029;

		WHILE_D6_TOGGLES(sa);

		bytecnt -= tot1;
	}

	READ_RESET();

    /* Now that the re-programming of flash is complete, reset: */
	{
#ifdef RESETMACRO
	RESETMACRO();
#else
    void    (*reset)();
    reset = RESETFUNC();
    reset();
#endif
	}

	return(0);	/* won't get here */

#else
	/* destination address must always be aligned to fdev->width,
	 * because that's how the hardware works */
	dest -= doft;

	do {
		/* the flash hardware forces us to write in
		 * address-aligned units of fdev->width; this loop
		 * gathers up an aligned unit of data from src, and
		 * writes it to dest; if such a unit isn't available,
		 * we fetch enough bytes from dest to pad out the
		 * difference */

		val = *(ftype *)dest;

		if (doft && doft != fdev->width)
		{
			/* we had to align dest above, so read bytes
			 * into dest buffer to account for that (this
			 * should happen only on the first iteration
			 * when dest was originally misaligned) */
			while (doft < fdev->width && bytecnt--)
				buf[doft++] = *src++;
		}
		else if (PTR_ALIGN(src, fdev->width) || bytecnt < fdev->width)
		{
			/* src is misaligned, or we have too few bytes
			 * to do an aligned-width transfer */
			for (wbyte = 0;
			     wbyte < fdev->width && bytecnt;
			     bytecnt--, wbyte++)
				buf[wbyte] = *src++;
		}
		else
		{
			/* src and dest are properly aligned, and we
			 * have enough bytes to do a full-width
			 * non-buffered transfer */
			val = *(ftype *)src;
			src += fdev->width;
			bytecnt -= fdev->width;
		}

		/* TODO: how to take advantage of write buffering?
		* One idea is to have the fdev structure indicate the
		* size of the buffer, and have the calling function
		* pre-align and organize everything into buffer-sized
		* chunks... */

		FLASH_WRITE(dest, buf);

		while (1)
		{
			WATCHDOG_MACRO;

			/* TODO: this is a clever way of checking for
			 * completion, but it assumes that the flash
			 * drops immediately back into read-array mode
			 * at the end of the programming cycle; not
			 * all chips do that (the s29's do, tho) */
			if (*(ftype *)dest == val
			    && *(ftype *)dest == val)
				break;

			/* TODO: I'm not quite sure the logic here is
			 * right; what if the flash chips don't finish
			 * at the same time, or one times out but the
			 * other doesn't? */
			if (D5_Timeout(dest)
			    && *(ftype*)dest != val) {
				goto done;
			}
		}

		dest += fdev->width;
	} while (bytecnt);

done:
	READ_RESET();
#endif

#ifdef RESETMACRO
	RESETMACRO();
#else
	{
		void    (*reset)();
		reset = RESETFUNC();
		reset();
	}
#endif

	return 0; /* never gets here */
}

/* EndS29gl512n_16x2_ewrite():
 * Function place holder to determine the end of the above function.
 */
void
EndS29gl512n_16x2_ewrite(void)
{}


/* S29gl512n_16x2_type():
 * Run the AUTOSELECT algorithm to retrieve the manufacturer and
 * device id of the flash.
 */
int
S29gl512n_16x2_type(struct flashinfo *fdev)
{
	ushort	man, dev;
	ulong	id;


	AUTO_SELECT();

	man = *(ftype *)(fdev->base);		/* manufacturer ID */
	dev = *(ftype *)((fdev->base + 56));	/* device ID */
	id = man;
	id <<= 16;
	id |= dev;

	fdev->id = id;

	READ_RESET();

	return((int)(fdev->id));
}

/* EndS29gl512n_16x2_type():
 * Function place holder to determine the end of the above function.
 */
void
EndS29gl512n_16x2_type(void)
{}

/**************************************************************************
 **************************************************************************
 *
 * The remainder of the code in this file should only included if the
 * target configuration is such that this device is the only real
 * flash device in the system that is to be visible to the monitor.
 *
 **************************************************************************
 **************************************************************************
 */
#ifdef SINGLE_FLASH_DEVICE

/* FlashXXXFbuf[]:
 * If FLASH_COPY_TO_RAM is defined then these arrays will contain the
 * flash operation functions above.  To operate on most flash devices,
 * you cannot be executing out of it (there are exceptions, but
 * in general, we do not assume the flash supports this).  The flash
 * functions are copied here, then executed through the function
 * pointers established in the flashinfo structure below.
 * One obvious requirement...  The size of each array must be at least
 * as large as the function that it will contain.
 */
#ifdef FLASH_COPY_TO_RAM
ulong    FlashTypeFbuf[400];
ulong    FlashEraseFbuf[400];
#ifdef BUFFERED_WRITE
ulong    FlashWriteFbuf[600];
#else
ulong    FlashWriteFbuf[400];
#endif
ulong    FlashEwriteFbuf[400];
#endif

/* FlashNamId[]:
 * Used to correlate between the ID and a string representing the name
 * of the flash device.
 */
struct flashdesc FlashNamId[] = {
	{ SPANSION_29GL512N_ID,		"Spansion-29GL512N" },
	{ 0, (char *)0 },
};


struct sectorinfo sinfo_bank0[SECTOR_TOTAL];
struct sectorinfo sinfo_bank1[SECTOR_TOTAL];

int
flashBankInit(int banknum, int snum, struct sectorinfo *sinfotbl)
{
	int		i;
	uchar	*begin;
	struct	flashinfo *fbnk;

	fbnk = &FlashBank[banknum];
	if (banknum == 0)
		fbnk->base = (unsigned char *)FLASH_BANK0_BASE_ADDR;
	else
		fbnk->base = (unsigned char *)FLASH_BANK0_BASE_ADDR + (SECTOR_TOTAL * SECTOR_SIZE);

	fbnk->width = 4;

#ifdef FLASH_COPY_TO_RAM
	fbnk->fltype = (int(*)())FlashTypeFbuf;
	fbnk->flerase = (int(*)())FlashEraseFbuf;
	fbnk->flwrite = (int(*)())FlashWriteFbuf;
	fbnk->flewrite = (int(*)())FlashEwriteFbuf;
#else
	fbnk->fltype = S29gl512n_16x2_type;
	fbnk->flerase = S29gl512n_16x2_erase;
	fbnk->flwrite = S29gl512n_16x2_write;
	fbnk->flewrite = S29gl512n_16x2_ewrite;
#endif

	/* This device doesn't support flash lock, so set the pointer
	 * to the default FlashLockNotSupported() function...
	 */
	fbnk->fllock = FlashLockNotSupported;

	fbnk->sectors = sinfotbl;
	fbnk->id = flashtype(fbnk);
	switch(fbnk->id) {
		case SPANSION_29GL512N_ID:
			fbnk->end = fbnk->base + (SECTOR_TOTAL * SECTOR_SIZE) - 1;
			fbnk->sectorcnt = SECTOR_TOTAL;
			break;
		default:
			/* The second bank may legitimately not be populated, so if it
			 * isn't found, indicate unavailable, don't claim an error...
			 */
			printf("Flash bank %d: ",banknum);
			if (banknum == 0)
				printf("invalid id: 0x%lx.\n",fbnk->id);
			else
				printf("not available.\n");
			return(-1);
	}

	begin = fbnk->base;
	for(i=0;i<fbnk->sectorcnt;i++,snum++) {
		fbnk->sectors[i].snum = snum;
		fbnk->sectors[i].size = SECTOR_SIZE;
		fbnk->sectors[i].begin = begin;
		fbnk->sectors[i].end = fbnk->sectors[i].begin + SECTOR_SIZE - 1;
		fbnk->sectors[i].protected = 0;
		begin += SECTOR_SIZE;
	}
	return(snum);
}

/* FlashInit():
 * Initialize data structures for each bank of flash...
 */
int
FlashInit()
{
	int	snum;

	FlashCurrentBank = 0;

	/* If code is in flash, then we must copy the flash ops to RAM.
	 * Note that this MUST be done when cache is disabled to assure
	 * that the RAM is occupied by the designated block of code.
	 */

#ifdef FLASH_COPY_TO_RAM
	if (flashopload((ulong *)S29gl512n_16x2_type,
		(ulong *)EndS29gl512n_16x2_type,
		FlashTypeFbuf,sizeof(FlashTypeFbuf)) < 0)
		return(-1);
	
	if (flashopload((ulong *)S29gl512n_16x2_erase,
		(ulong *)EndS29gl512n_16x2_erase,
		FlashEraseFbuf,sizeof(FlashEraseFbuf)) < 0)
		return(-1);
	
	if (flashopload((ulong *)S29gl512n_16x2_ewrite,
		(ulong *)EndS29gl512n_16x2_ewrite,
		FlashEwriteFbuf,sizeof(FlashEwriteFbuf)) < 0)
		return(-1);
	
	if (flashopload((ulong *)S29gl512n_16x2_write,
		(ulong *)EndS29gl512n_16x2_write,
		FlashWriteFbuf,sizeof(FlashWriteFbuf)) < 0)
		return(-1);
#endif

#ifdef IS_MONOLITHIC_FLASH
	ActiveFlashBanks = 1;
	snum = flashBankInit(0,0,sinfo_bank0);
	if (!IS_MONOLITHIC_FLASH()) {
		if (flashBankInit(1,snum,sinfo_bank1) != -1)
			ActiveFlashBanks++;
	}
#else
	snum = flashBankInit(0,0,sinfo_bank0);
#endif
	if (snum < 0)
		return(-1);

#ifdef FLASH_PROTECT_RANGE
	sectorProtect(FLASH_PROTECT_RANGE,1);
#endif

#if FLASHRAM_BASE
	FlashRamInit(snum, FLASHRAM_SECTORCOUNT, &FlashBank[FLASHRAM_BANKNUM],
		sinfoRAM, 0);
#endif

	return(0);
}

#endif	/* SINGLE_FLASH_DEVICE */

#endif	/* INCLUDE_FLASH */
