/* reset.s:
 *
 * First bit of boot code run by the processor.
 *
 */
 #include "config.h"

    .file   "reset.s"

    .extern start
    .extern MonStack
    
    .extern LedSet
    .extern LedClr
    .extern LedClrAll

    /*
     * Linker defined symbols used to initialise bss.
     */

    .globl __bss_start
    .globl __bss_end


     /*
      * Explicitly allow the use of r1 (the assembler temporary register)
      * within this code. This register is normally reserved for the use of
      * the compiler.
      */

    .set noat

/*********************************************************************
 *
 * _hardware_reset
 *
 * This is the reset entry point for Nios II.
 *
 * At reset, only the cache line which contain the reset vector is
 * initialized. Therefore the code within the first cache line is required
 * to initialize the instruction cache for the remainder of the code.
 *
 * Note that the automatically generated linker script requires that
 * the .init section is less than 0x20 bytes long.
 *
 */

    .section .entry, "xa"
    .align    2
    .global   reset

reset:


    /* Initialize all cache lines within the instruction cache */

#if NIOS2_ICACHE_SIZE > 0

    /*
     * The assumption here is that the instruction cache size is always
     * a power of two.
     */

#if NIOS2_ICACHE_SIZE > 0x8000
    movhi r1, %hi(NIOS2_ICACHE_SIZE)
#else
    movui r1, NIOS2_ICACHE_SIZE
#endif
    movui r2, NIOS2_ICACHE_LINE_SIZE

0:
    initi r1
    sub   r1, r1, r2
    bgt   r1, zero, 0b

#endif /* NIOS2_ICACHE_SIZE > 0 */

    /*
     * Having initialised the cache, call the entry point for the .text
     * section.
     */

    movhi r1,     %hiadj(coldstart)
    addi  r1, r1, %lo(   coldstart)

    jmp   r1

    .section .text
    .align 2



#include "warmstart.h"

/* Depending on the CPU, these entries may have to be shuffled
 * around a bit.  The position of coldstart is very CPU specific
 * and the location of moncomptr should be in a place where it is
 * unlikely to be moved if other monitor functionality is changed.
 */

/*********************************************************************
 *
 * coldstart:
 * The reset point.
 */
    .section .text
    .align    2
    .global   coldstart
    .type     coldstart, @function

coldstart:
    /* - Invalidate and disable caches
     * - Load INITIALIZE (defined in warmstart.h) into whereever it
     *   needs to be put so that warmstart sees it as an argument
     *   passed to it.
     *
     * - Jump to warmstart
     */

    /* clear all control registers */
    wrctl status ,  zero  //
    wrctl estatus,  zero
    wrctl bstatus,  zero
    wrctl ienable,  zero

    /*
     * Clear bss.
     *
     * This uses the symbols: __bss_start and __bss_end, which are defined
     * by the linker script. They mark the begining and the end of the bss
     * region. The linker script guarantees that these values are word
     * aligned.
     */

    movhi r1, %hiadj( __bss_start)
    addi  r1, r1, %lo(__bss_start)

    movhi r2, %hiadj( __bss_end)
    addi  r2, r2, %lo(__bss_end)

    beq r1, r2, 3f

2:
    stw zero, (r1)
    addi r1, r1, 4
    bltu r1, r2, 2b

3:


     br warmstart

/*********************************************************************
 *
 * moncomptr:
 * Pointer to the moncom function, used to link application to monitor.
 * Refer to umon_main/target/core/moncomptr.S
 */
#include "moncomptr.S"

/*********************************************************************
 *
 * etheraddr:
 * Location that could be used to store a fixed MAC address.
 * Refer to umon_main/target/core/etheraddr.S
 */
#include "etheraddr.S"

/*********************************************************************
 *
 * warmstart:
 * A point callable by C, as warmstart(int type), where
 * 'type' is one of the values defined in warmstart.h.
 */
    .global   warmstart
    .section .text
    .align    2
    .type     warmstart, @function

warmstart:
    /* Minimal CPU/IO Initialization (i.e. chip selects) here.
     * Stay in assembler here if possible.
     */

#if NIOS2_DCACHE_SIZE > 0

    /*
     * The assumption here is that the instruction cache size is always
     * a power of two.
     */

#if NIOS2_DCACHE_SIZE > 0x8000
    movhi r1, %hi(NIOS2_DCACHE_SIZE)
#else
    movui r1, NIOS2_DCACHE_SIZE
#endif

    movui r2, NIOS2_DCACHE_LINE_SIZE

1:
    initi r1
    sub   r1, r1, r2
    bgt   r1, zero, 1b

#endif /* NIOS2_DCACHE_SIZE > 0 */


/*********************************************************************
 *
 * SPInit:
 * Establish a stack frame.
 */
SPInit:
    /* First initialize stack to point to MonStack+MONSTACKSIZE+64
     * (an address that is outside the MonStack array), and then
     * call stkinit().  This loads the MonStack[] array with a known
     * pattern that allows uMon to later analyze the running stack usage.
     */

     movhi sp,  %hiadj(MonStack + MONSTACKSIZE + 64)
     addi  sp, sp, %lo(MonStack + MONSTACKSIZE + 64)

     call  stkinit

    /* Next, re-initialize stack to point to MonStack+MONSTACKSIZE...
     * This is important because other portions of the code
     * assume this is where the stack resides.
     */
     movhi sp,  %hiadj(MonStack + MONSTACKSIZE - 4)
     addi  sp, sp, %lo(MonStack + MONSTACKSIZE - 4)

/*********************************************************************
 *
 * gotoC:
 * This code jumps to the start() function in the monitor
 * and should never return.
 */
gotoC:
    /* - Retrieve the variable passed to warmstart and place it
     *   whereever it needs to be so that the start() function in
     *   'C' sees it as an argument.  Note that if FORCE_BSS_INIT
     *   is defined, then start() will ignore this argument.
     * - Branch to start().
     */

  
     movi   r4, INITIALIZE
     call   start


/*********************************************************************
 *
 * _exception_vector
 *
 * This is the exception entry point. It is responsible for determing if
 * the exception was caused by an hardware interrupt, or a software exception.
 * It then vectors accordingly using the a table to handle the exception.
 * This needs to be located at the exception address specified at system
 * generation time, usually within SOPC Builder "more "cpu" settings".
 *
 * This routine depends on an externally declared vsr_table[2] ( vector
 * service routine).  Entry [0] contains the address of the hardware
 * exception handler and entry [1] contains the address of the software
 * exception handler.
 *
 */

    .section .irq, "xa"

_exception_vector:

    /*
     * First, re-wind so we're pointed to the instruction where the exception
     * occurred.
     */

    addi ea, ea, -4

    /*
     * Now test to determine the cause of the exception.
     */

    rdctl r24, estatus
    andi  r24, r24, 1
    beq   r24, zero, 4f
    rdctl r24, ipending
    beq   r24, zero, 4f

    /*
     * Falling through to here means that this was a hardware interrupt.
     * Determine if we call the user specified handler or umon's
     */

    movi  r24, 4

 4:
    /*
     * r24 now contains the byte offset of the exception handler
     * in the VSR table
     */

    .extern hal_vsr_table

    /*
     * load the address of the excetion handler, and jump to it
     */

    orhi r24, r24, %hiadj(hal_vsr_table)
    addi r24, r24, %lo(   hal_vsr_table)
    ldw  r24,(r24)
    jmp  r24
