#include <stdint.h>
#include "config.h"
#include "stddefs.h"
#include "genlib.h"
#include "cache.h"
#include "warmstart.h"
#include "cpuio.h"
#include "timer.h"
#include "cdefBF537.h"
#include "date.h"
#include "i2c.h"

/* setTimeAndDate() & getTimeAndDate():
 * These two functions interface to the portion of the RAMTRON device
 * that deals with maintaining time-of-day...
	char day;		 1-7 (sun = 1)
	char date;		 1-31
	char month;		 1-12
	char hour;		 0-23
	char minute;	 0-59
	char second;	 0-59
	short year;		 full number (i.e. 2006)
 */

#if HARDWARE_IS_REVA

#define W_BIT				0x02
#define R_BIT				0x01
#define OSCEN_BIT			0x80

int
getTimeAndDate(struct todinfo *now)
{
	int rc;
	uchar rbuf[25], wbuf[2];

	/* Read register 0 so that we can modify the 'W' bit without
	 * affecting any other bits...
	 */
	rbuf[0] = 1; rbuf[1] = 0;
	if ((rc = i2cRead(0,RAMTRON_COMPANION | REPEATED_START,rbuf,1)) != 0)
		return(-1);

	/* Clear the 'R' bit...
	 */
	wbuf[0] = 0; wbuf[1] = rbuf[0] &= ~R_BIT;
	if ((rc = i2cWrite(0,RAMTRON_COMPANION,wbuf,2)) != 0)
		return(-1);

	/* Set the 'R' bit...
	 */
	wbuf[0] = 0; wbuf[1] = rbuf[0] | R_BIT;
	if ((rc = i2cWrite(0,RAMTRON_COMPANION,wbuf,2)) != 0)
		return(-1);

	/* Read 8 bytes of data out of RAMTRON's CPU Companion...
	 */
	rbuf[0] = 1; rbuf[1] = 2;
	if ((rc = i2cRead(0,RAMTRON_COMPANION | REPEATED_START,rbuf,7)) != 0)
		return(-1);
	
	now->year = 2000 + (((rbuf[6] & 0xf0) >> 4) * 10) + (rbuf[6] & 0x0f);
	now->month = (((rbuf[5] & 0x10) >> 4) * 10) + (rbuf[5] & 0x0f);
	now->date = (((rbuf[4] & 0x30) >> 4) * 10) + (rbuf[4] & 0x0f);
	now->day = (rbuf[3] & 0x07);
	now->hour = (((rbuf[2] & 0x30) >> 4) * 10) + (rbuf[2] & 0x0f);
	now->minute = (((rbuf[1] & 0x70) >> 4) * 10) + (rbuf[1] & 0x0f);
	now->second = (((rbuf[0] & 0x70) >> 4) * 10) + (rbuf[0] & 0x0f);
	return(0);
}

int
setTimeAndDate(struct todinfo *now)
{
	int rc;
	uchar wbuf[16], rbuf[8];

	/* Read registers 0&1 so that we can modify the 'W' bit (in reg0)
	 * and the OSCEN bit (in reg1) without affecting any other bits...
	 */
	rbuf[0] = 1; rbuf[1] = 0;
	if ((rc = i2cRead(0,RAMTRON_COMPANION | REPEATED_START,rbuf,2)) != 0)
		return(-1);

	/* If the incoming pointer is 0, this is a request to disable
	 * the oscillator...
	 */
	if (now == 0) {
#if 0
		wbuf[0] = 1; wbuf[1] = rbuf[1] | OSCEN_BIT;
		if ((rc = i2cWrite(0,RAMTRON_COMPANION,wbuf,2)) != 0)
			return(-1);
#else
		printf("Oscillator can't be turned off by this command.\n");
#endif
		return(0);
	}

	/* Set 'W' bit high to freeze the clock while writing...
	 */
	wbuf[0] = 0; wbuf[1] = rbuf[0] | W_BIT;
	if ((rc = i2cWrite(0,RAMTRON_COMPANION,wbuf,2)) != 0)
		return(-1);

	/* Build the buffer to be sent to RAMTRON...
	 * Registers 1-8 (inclusive) are affected.
	 * Reg1 has teh OSCEN bit, and regs2-8 have the time/date.
	 */
	wbuf[0] = 1;	/* First byte of the burst is the register offset */
	wbuf[1] = rbuf[1] & ~OSCEN_BIT;
	wbuf[2] = (((now->second/10) << 4) | (now->second % 10));
	wbuf[3] = (((now->minute/10) << 4) | (now->minute % 10));
	wbuf[4] = (((now->hour/10) << 4) | (now->hour % 10));
	wbuf[5] = now->day & 7;
	wbuf[6] = (((now->date/10) << 4) | (now->date % 10));
	wbuf[7] = (((now->month/10) << 4) | (now->month % 10));
	while(now->year > 99) now->year -= 100;
	wbuf[8] = (((now->year/10) << 4) | (now->year % 10));
	if ((rc = i2cWrite(0,RAMTRON_COMPANION,wbuf,9)) != 0)
		return(-1);

	/* Clear 'W' bit to release the clock now that the write is done...
	 */
	wbuf[0] = 0; wbuf[1] = rbuf[0] & ~W_BIT;
	if ((rc = i2cWrite(0,RAMTRON_COMPANION,wbuf,2)) != 0)
		return(-1);
	
	return(0);
}

#elif HARDWARE_IS_REVB

/* Much of this code is snipped from MAXIM Appnote 517 titled:
 *
 * "DS1371/DS1372/DS1374 32-Bit Binary Counter Time Conversion"
 *
 */
struct tm_struct {
	int tm_sec;		/* seconds after the minute - [0,59] */
	int tm_min;		/* minutes after the hour - [0,59] */
	int tm_hour;	/* hours since midnight - [0,23] */
	int tm_mday;	/* day of the month - [1,31] */
	int tm_mon;		/* months since January - [0,11] */
	int tm_year;	/* years since 1900 */
	int tm_wday;	/* days since Sunday - [0,6] */
	int tm_yday;	/* days since January 1 - [0,365] */
};

unsigned long DaysToMonth[13] = {
	0,31,59,90,120,151,181,212,243,273,304,334,365
};

unsigned long
DS1371_DateToBinary(struct tm_struct *dt)
{
	unsigned long iday, val, dtm;

	dtm = DaysToMonth[dt->tm_mon];
	iday = 365 * (dt->tm_year - 70) + dtm + (dt->tm_mday - 1);
	iday = iday + (dt->tm_year - 69) / 4;
	if ((dt->tm_mon > 1) && ((dt->tm_year % 4) == 0)) {
		iday++;
	}
	val = dt->tm_sec + 60 * dt->tm_min + 3600 * (dt->tm_hour + 24 * iday);

	return val;
}

void
DS1371_BinaryToDate(unsigned long binary,struct tm_struct *dt)
{
	unsigned long days_to_month, day_of_week;
	unsigned long hour, day, minute, second, month, year;
	unsigned long whole_days_since_1968, leap_year_periods;
	unsigned long days_since_current_lyear, days_since_first_of_year;
	unsigned long whole_minutes, whole_hours, whole_days, whole_years;

	whole_minutes = binary / 60;
	second = binary - (60 * whole_minutes);			// leftover seconds
	whole_hours = whole_minutes / 60;
	minute = whole_minutes - (60 * whole_hours);	// leftover minutes
	whole_days = whole_hours / 24;
	hour = whole_hours - (24 * whole_days);			// leftover hours
	whole_days_since_1968 = whole_days + 365 + 366;
	leap_year_periods = whole_days_since_1968 / ((4 * 365) + 1);
	days_since_current_lyear = whole_days_since_1968 % ((4 * 365) + 1);

	// if days are after a current leap year then add a leap year period
	if ((days_since_current_lyear >= (31 + 29))) {
		leap_year_periods++;
	}
	whole_years = (whole_days_since_1968 - leap_year_periods) / 365;
	days_since_first_of_year =
		whole_days_since_1968 - (whole_years * 365) - leap_year_periods;

	if ((days_since_current_lyear <= 365) && (days_since_current_lyear >= 60)) {
		days_since_first_of_year++;
	}
	year = whole_years + 68;

	// setup for a search for what month it is based on how many days
	// have past within the current year
	month = 13;
	days_to_month = 366;
	while (days_since_first_of_year < days_to_month) {
		month--;
		days_to_month = DaysToMonth[month];
		if ((month >= 2) && ((year % 4) == 0)) {
			days_to_month++;
		}
	}
	day = days_since_first_of_year - days_to_month + 1;
	day_of_week = (whole_days + 4) % 7;

	dt->tm_yday =
		days_since_first_of_year;	// days since January 1 - [0,365]
	dt->tm_sec = second;			// seconds after the minute - [0,59]
	dt->tm_sec = second;			// seconds after the minute - [0,59]
	dt->tm_min = minute;			// minutes after the hour - [0,59]
	dt->tm_hour = hour;				// hours since midnight - [0,23]
	dt->tm_mday = day;				// day of the month - [1,31]
	dt->tm_wday = day_of_week;		// days since Sunday - [0,6]
	dt->tm_mon = month;				// months since January - [0,11]
	dt->tm_year = year;				// years since 1900
}

int
getTimeAndDate(struct todinfo *now)
{
	int rc;
	uchar *cp, rbuf[16];
	unsigned long bval;
	struct tm_struct tm;

	rbuf[0] = 1; rbuf[1] = 0;
	if ((rc = i2cRead(0,MAXIM_RTC | REPEATED_START,rbuf,10)) != 0)
		return(-1);

	cp = (uchar *)&bval;
	cp[0] = rbuf[0];
	cp[1] = rbuf[1];
	cp[2] = rbuf[2];
	cp[3] = rbuf[3];

	DS1371_BinaryToDate(bval,&tm);

	/* Convert to todinfo structure...
	 */
	now->year = tm.tm_year + 1900;
	now->month = tm.tm_mon + 1;
	now->date = tm.tm_mday;
	now->day = tm.tm_wday;
	now->hour = tm.tm_hour;
	now->minute = tm.tm_min;
	now->second = tm.tm_sec;
	
	return(0);
}

int
setTimeAndDate(struct todinfo *now)
{
	int rc;
	uchar *cp, wbuf[16];
	unsigned long bval;
	struct tm_struct tm;

	/* Convert from todinfo to tm_struct structure...
	 */
	tm.tm_year = now->year - 1900;
	tm.tm_mon = now->month - 1;
	tm.tm_mday = now->date;
	tm.tm_wday = now->day;
	tm.tm_hour = now->hour;
	tm.tm_min = now->minute;
	tm.tm_sec = now->second;
	
	bval = DS1371_DateToBinary(&tm);
	cp = (uchar *)&bval;
	wbuf[0] = 0;
	wbuf[1] = cp[0];
	wbuf[2] = cp[1];
	wbuf[3] = cp[2];
	wbuf[4] = cp[3];

	rc = i2cWrite(0,MAXIM_RTC,wbuf,5);
	return(0);
}

#endif
