/* cpuio.c:
 * This file contains the generic serial port drivers and other
 * miscellaneous target specific functions.
 *
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 *
 * Author:  Ed Sutter
 * email:   esutter@lucent.com
 * phone:   908-582-2351
 */
#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "stddefs.h"
#include "ppcuic.h"
#include "uart16550.h"
#include "flash.h"

extern int i2c_init(void);
extern int fs6377_init(int mode);
extern int ps2_kbd_init(void);
extern int ps2_gotachar(void);
extern uchar ps2_getchar(void); 
extern void sed_pwr_dn(void);
extern int sed135x_init(void);
extern void sed_lcd_bkl(uchar bright);
extern void sed_putchar(char c);
extern int sed_color_depth;
extern int sed135x_ok;

// We use the SED GPIO1 to control the LCD backlight
void lcd_bkl(uchar bright)
{
#if INCLUDE_SED
	sed_lcd_bkl(bright);
#endif
}

int
getUartDivisor(int baud, unsigned char *hi, unsigned char *lo)
{
	*lo =  (230400/baud); // 3.6864Mhz/16 = 230400
	*hi =  0;
	return(0);
}

int
devInit(int baud)
{
	extern void sed135x_on();

	// Initialize the UART
    InitUART(baud);

	// Initialize the I2C Port
	i2c_init();

#if INCLUDE_SED
	sed_pwr_dn();
#endif

	// Initialize the FS6377 Clock Generator assuming
	// CRT Mode for now.  If need be, it will be changed
	// to the correct value later when the display is
	// initialized.  We do this now to get the serial
	// clock running
	fs6377_init(SED_DISP_MODE_CRT);

#if INCLUDE_SED
	// Initialize the display controller
	sed_color_depth = 4;
	sed135x_init();
	if (sed135x_ok == 0)
		printf("sed135x_init() failed!\n");
	sed135x_on();
#endif

#if INCLUDE_PS2KBD
	// Initialize the PS2 controller
	ps2_kbd_init();
#endif
	
	return 0;
}

void
initCPUio()
{
	ppcMtuicsr(0xFFFFFFFF);		/* clear all ints */
	ppcMtuicer(0x00000000);		/* disable all ints */
	ppcMtuiccr(0);				/* Set all ints to non-critical. */
	ppcMtuicpr(0xFFFFE07F);		/* Set up positive interrupt polarity. */
	ppcMtuicsr(0xFFFFFFFF);		/* clear all ints */
}

void
cacheInitForTarget()
{
    extern  void ppcMticcr();

    ppcMticcr(0x80000001);
}

/* uSleep():
 * Reasonably accurate microsecond-resolution delay function hooked to
 * the Low Time Base register of the PPC405.  It assumes the time base
 * is derived from the CPU clock, so CPC0_CR1[CETE] should be low.
 * Example usage of GNU-only syntax for using a C-variable in a
 * line of assembler:
 *  read:
 *  asm volatile ("mfdcr %0, 0xa0" : "=r" (tmp)); 
 *  write:
 *  asm volatile ("mtdcr 0xa0, %0" : "=r" (tmp) : "0" (tmp));
 */
void
uSleep(int delay)
{
	unsigned duration;
	unsigned register begin, end;

	duration = delay * MON_CPU_CLOCK;

	/* Retrieve value from TBL first time...
	 */
	asm volatile ("mfspr %0, 0x10c" : "=r" (begin)); /* TBL */

	/* Poll TBL until the delta is 1/1000th of a second...
	 */
	do {
		asm volatile ("mfspr %0, 0x10c" : "=r" (end)); /* TBL */
	} while((end - begin) < duration);
}

void
flash_lock(void)
{
    char    cmd[32];

	/* This hardware has hardware flash sector protection, so use it...
	 * This means that a "newmon" must be preceded by a "flash unlock"
	 * to successfully erase and rewrite the protected boot sectors.
	 */
	sprintf(cmd,"flash lock %s",FLASH_PROTECT_RANGE);
	docommand(cmd,1);
}
