    .file "rom_reset.s"

/*
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 *
 * Author:  Ed Sutter
 * email:   esutter@lucent.com
 * phone:   908-582-2351
 * 
 *
 * Modified for the CSB335 - LH79520 (ARM720T) Single Board
 *
 * rom_reset.s:
  */

#include "config.h"
#include "warmstart.h"
#include "lh79520.h"

    /*
     * Have a separate stack for each processor mode.
     */

    /* define sizes for each mode's stack */
    .equ FiqStackSz, 4096
    .equ IrqStackSz, 4096
    .equ AbtStackSz, 4096
    .equ UndStackSz, 4096
    .equ SysStackSz, 4096

    /* declare the stacks */
    .extern MonStack
    .global FiqStack
    .global IrqStack
    .global AbtStack
    .global UndStack
    .global SysStack        
    
    /* allocate the stacks */
    .comm   FiqStack, FiqStackSz    /* for the FIQ mode */
    .comm   IrqStack, IrqStackSz    /* for the IRQ mode */
    .comm   AbtStack, AbtStackSz    /* for the Abort mode */
    .comm   UndStack, UndStackSz    /* for the Undef mode */
    .comm   SysStack, SysStackSz    /* for the System mode */
    /* User mode has the same stack as system mode. */

/*********************************************************************/
    
    .extern start

    .global reset
    .global coldstart
    .global lukewarmstart
    .global warmstart
    .global etheraddr
    .global moncomptr

    .text
        
    /*
     * Exception table at address 0
     */ 
reset:  
    b coldstart
    b undefined_instruction
    b software_interrupt
    b abort_prefetch
    b abort_data
    b not_used
    b interrupt_request
    b fast_interrupt_request

/********************************************************************/

    /* Provide space to allow a programmer to place an ASCII 
     * string in this location as an optional point
     * of storage for MAC address...
     */ 
etheraddr:
    .byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
    .byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
    .byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff

moncomptr:
    .long   moncom

undefined_instruction:
	b undefined_instruction

software_interrupt:
	b software_interrupt

abort_prefetch:
	b abort_prefetch

abort_data:
	b abort_data

not_used:
	b not_used

interrupt_request:
	b interrupt_request

fast_interrupt_request:
	b fast_interrupt_request

/*********************************************************************/
coldstart:

    /*
     * At the end of the reset sequence, MMU, Icache, Dcache,
     * and write buffer are all disabled.
     * Also IRQs and FIQs are disabled in the processor's CPSR
     * The operating mode is SVC (supervisory mode), and the
     * PC is vectored at 0x00000000. A branch in 0x00000000
     * brings us directly here.
	 *
	 */

	ldr	pc, =coldstart_1
	nop

coldstart_1:
	ldr r0, =0x2001					/* allow access to all coprocessors */
	mcr 	p15,0,r0,c15,c1,0
	nop
	nop
	nop

	ldr 	r0, =0x00000078
	mcr 	p15,0,r0,c1,c0,0		/* Disable MMU, caches, write buffer */
	nop
	nop
	nop

	ldr 	r0, =0x00000000
	mcr     p15,0,r0,c8,c7,0   		/* flush TLB's */
	mcr     p15,0,r0,c7,c7,0   		/* flush Caches */
	mcr		p15,0,r0,c7,c10,4		/* Flush Write Buffer */
	nop
	nop
	nop

	mvn    r0, #0                   /* grant manager access to all domains */
	mcr    p15,0,r0,c3,c0,0  

    /* Mask off all IRQs and FIQs */
    ldr r2, =0xffffffff	        	/* clear all bits */
    ldr r1,	=0xfffff014
    str r2, [r1]

	/*
     * Initialize the LH79520 Clocks
     */
	// unlock RPC and set HCLK as CLKOUT
    ldr r2, =0x00000263
    ldr r1, =0xfffe2000
    str r2, [r1]

	// set FCLK to PLL/4 (77.41Mhz)
	ldr r2, =0x00000002
	ldr r1, =0xfffe201c
    str r2, [r1]

#ifdef INIT_SDRAM
	// set HCLK to PLL/6 (51.61Mhz) - This also sets the SDRAM clock
	// and the staic memory clock
	ldr r2, =0x00000003
	ldr r1, =0xfffe2018
    str r2, [r1]
#endif

	// enable peripheral clocks
	ldr r2, =0x00000000
	ldr r1, =0xfffe2024
    str r2, [r1]
	ldr r1, =0xfffe2028
    str r2, [r1]

	// enable sdram clocks
	ldr r2, =0x00000000
	ldr r1, =0xfffe202c
    str r2, [r1]

	// enable all memory pins
	ldr r2, =0xffffffff
	ldr r1, =0xfffe5000
    str r2, [r1]

	// enable uart0 and uart1 and ssp pins
	ldr r2, =0x0000000f
	ldr r1, =0xfffe5010
    str r2, [r1]

	ldr r2, =0x0000001c
	ldr r1, =0xfffe5014
    str r2, [r1]

	// enable clkout and nWAIT
	ldr r2, =0x00000502
	ldr r1, =0xfffe5008
    str r2, [r1]

	// Setup nCS0 - StrataFLASH, 16Mbytes, 200nsec, 16-btis
	ldr r2, =0x13002ca1
	ldr r1, =0xffff1000
	str r2, [r1]


	// Setup nCS1 - LAN91C113 - 16-bit external ready
	ldr r2, =0x13001c60
	ldr r1, =0xffff1004
	str r2, [r1]


	/* pause for ~200usec for clocks to stablize
	bl	delay_200

    /*
     * Initialize SDRAM Controller
     */
#ifdef INIT_SDRAM
	// Setup SDRAM
	// issue 2 NOP
    ldr r2, =0x00000003
    ldr r1, =0xffff2004
    str r2, [r1]
	bl	delay_200
    str r2, [r1]
	bl	delay_200

	// issue Precharge All
    ldr r2, =0x00000001
    ldr r1, =0xffff2004
    str r2, [r1]
	bl	delay_200

	// set refresh timer to 16 clocks to a force bunch right away
    ldr r2, =0x00000010
    ldr r1, =0xffff2008
    str r2, [r1]
	bl	delay_200

	// now set to the final number to ~15usec @ 51Mhz
    ldr r2, =0x000002e0
    ldr r1, =0xffff2008
    str r2, [r1]
	bl	delay_200

	// set MRS mode
    ldr r2, =0x00000002
    ldr r1, =0xffff2004
    str r2, [r1]
	bl	delay_200

	// 0x22 sets burst mode and cas latency for bank 0
    ldr r1, =0x20022000
    ldr r2, [r1]
	bl	delay_200

	// Write SDRCConfig0, for CAS latency = 2, RAS to CAS = 2, 32-bit, 4 bank
    ldr r2, =0x01a00088
    ldr r1, =0xffff2000
    str r2, [r1]
	bl	delay_200
#endif

	// set Normal mode
    ldr r2, =0x00000000
    ldr r1, =0xffff2004
    str r2, [r1]
	bl	delay_200

/********************************************************************/

midstart:
    ldr r0, =INITIALIZE
    
    /* fall-through to 'lukewarmstart' */

/********************************************************************/
    
lukewarmstart:  
    /* Save the argument to r11 */
    mov r11, r0

    /*
     * *** DO NOT TOUCH R11 ***
     */ 

    /*
     * Set-up the stack-pointers for all operating modes
     */

    /* FIQ mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x11           /* set FIQ mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(FiqStack + FiqStackSz - 4)    /* initialize the stack ptr */
    /* IRQ mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x12           /* set IRQ mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(IrqStack + IrqStackSz - 4)    /* initialize the stack ptr */
    /* Abort mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x17           /* set Abort mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(AbtStack + AbtStackSz - 4)    /* initialize the stack ptr */
    /* Undef mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x1b           /* set Undef mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(UndStack + UndStackSz - 4)    /* initialize the stack ptr */
    /* System mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x1f           /* set System mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    ldr sp, =(SysStack + SysStackSz - 4)    /* initialize the stack ptr */
    /* 'warmstart' will take us back to SVC mode
       stack for SVC mode will also be setup in warmstart */

    mov r0, r11     /* get argument back from r11 */
    b   warmstart

    
/********************************************************************/
    
warmstart:
    /* Save the argument to r11 */
    mov r11, r0

    /*
     * *** DO NOT TOUCH R11 ***
     */ 


    /* Change (back) to SVC mode */
    mrs r0, cpsr                /* move CPSR to r0 */
    bic r0, r0, #0x1f           /* clear all mode bits */
    orr r0, r0, #0x13           /* set System mode bits */
    msr CPSR_c, r0              /* move back to CPSR */
    /* Reset the stack pointer for the SVC mode (our current mode) */
    ldr sp, =(MonStack + MONSTACKSIZE - 4)

    /*
     * Restore argument which was saved to r11 and jump to
     * the C function start().
     */ 
    
    mov r0, r11
jump_to_c:
    bl start

    /* the C code should never return */
    b reset

.align 4


//********************************************************************/
// simple delay loop
delay_200:
	ldr		r3, =200			/* loop count */
delay_loop:
	subs	r3,r3,#1
	bne		delay_loop
	nop

	mov		pc, lr
