//==========================================================================
//
// csb337_gpio.h
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Contributors: 
// Date:         6/20/2003
// Description:  This file contains the board level to vitual
//				 GPIO bit assignments
//

// The AT91RM9200 GPIO's are spread across four 32-bit ports A-D.
// For bits that have an alternate function assigned, we use the bit 
// position within the respective port as we need that to setup the
// registers.  But for bits that are used as GPIOs we assign a 
// virtual GPIO number which is used by the GPIO functions
//
// This section also defines the bit setting for the alternate
// functions of the pins.  If we want alternate function A, we prepend
// an A, for function B, a B.  This insures we will get an error, or
// at least it will be very visible if we try to write assign a function
// to the wrong side.  Note the A or B is not related to the Port A
// B, just the alternate function.
//
// Finally we build the default values to be written into the GPIO registers
// in gpio_init (in csb337_gpio.c).


#ifndef CSB337_GPIO_H
#define CSB337_GPIO_H

#include "bits.h"

// PORT A							   BIT	A/B	DIR	SENSE	DESCRIPTION
#define A_SPI_DIN		BIT0		// PA0  A	-	-		SPI Data in
#define	A_SPI_DOUT		BIT1		// PA1  A	-	-		SPI Data Out
#define A_SPI_CLK		BIT2		// PA2  A	-	-		SPI Clock
#define A_SPI_CS0		BIT3		// PA3  A	-	-		SPI Chip Select 0 - CAN 
#define A_SPI_CS1		BIT4		// PA4  A	-	-		SPI Chip Select 1 - Expansion
#define PA5			BIT5			// PA5  A	-	-		Expansion GPIO/SPI_CS2/USART 3 TXD
#define PA6			BIT6			// PA6  - 	-	-		Expansion GPIO/SPI_CS3/USART 3 RXD
#define A_ETXCK			BIT7		// PA7  A	-	-		Ethernet Transmit Clock
#define A_ETXEN			BIT8		// PA8  A	-	-		Ethernet Transmit Enable
#define A_ETXD0			BIT9		// PA9  A	-	-		Ethernet Transmit Data 0
#define A_ETXD1			BIT10		// PA10 A	-	-		Ethernet Transmit Data 1
#define A_ECRS			BIT11		// PA11 A	-	-		Ethernet Carrier Sense 
#define A_ERXD0			BIT12		// PA12 A	-	-		Ethernet Receive Data 0
#define A_ERXD1			BIT13		// PA13 A	-	-		Ethernet Receive Data 1
#define A_ERXER			BIT14		// PA14 A	-	-		Ethernet Receive Error
#define A_EMDC			BIT15		// PA15 A	-	-		Ethernet MD Clock
#define A_EMDIO			BIT16		// PA16 A	-	-		Ethernet MD Data
#define PA17			17			// PA17 - 	In	-		Expansion GPIO
#define PA18			18			// PA18 - 	In	-		Expansion GPIO
#define PA19			19			// PA19 - 	In	-		Expansion GPIO
#define PA20			20			// PA20 - 	In	-		Expansion GPIO
#define USR_SW2			21 			// PA21 - 	In	Low		User Switch 2
#define PA22			22 			// PA22 - 	In	Low		Expansion GPIO/USART 2 TXD
#define PA23			23 			// PA23 - 	In	Low		Expansion GPIO
#define USB_SCON		24			// PA24 - 	In	High	USB Soft Connect
#define SDA				25			// PA25 - 	I/O	-		I2C Data - GPIO mode
#define SCL				26 			// PA26 - 	I/O	-		I2C Clock - GPIO mode
#define A_SDCK			BIT27		// PA27 A 	- 	- 		SD Card Clock 
#define A_SDCMD			BIT28		// PA28 A 	- 	- 		SD Card Command
#define A_SDD0			BIT29		// PA29 A 	- 	- 		SD Card Data 0
#define A_DBG_RXD		BIT30 		// PA30 A 	- 	- 		Debug Receive
#define A_DBG_TXD		BIT31 		// PA31 A 	- 	-		Debug Transmit

// default settings for PORT A
// Peripheral A select - SPI, SD Card 0, Ethernet and Debug UART
#define DEF_PIOA_ASR 	(A_SPI_DIN | A_SPI_DOUT | A_SPI_CLK | A_SPI_CS0 | A_SPI_CS1 | A_ETXCK | A_ETXEN | A_ETXD0 | A_ETXD1 | A_ECRS | A_ERXD0 | A_ERXD1 | A_ERXER | A_EMDC | A_EMDIO | A_SDCK | A_SDCMD | A_SDD0 | A_DBG_RXD | A_DBG_TXD)

// Peripheral B select - None
#define DEF_PIOA_BSR	0	

 // enable pullups where desired - usually just expansion GPIO
#define DEF_PIOA_PUER	(PA5 | PA6 | PA17 | PA18 | PA19 | PA20 | PA22 | PA23)

// enable default state of outputs that need to start high
#define DEF_PIOA_SODR	0

// enable default state of outputs that need to start low
#define DEF_PIOA_CODR	0

// enable all pins that start as outputs
#define DEF_PIOA_OER	0

// PORT B							   BIT	A/B	DIR	SENSE	DESCRIPTION
#define USR_LED2 		32			// PB0  -	Out Low		User LED 2
#define USR_LED1 		33			// PB1  -	Out Low		User LED 1
#define USR_LED0 		34			// PB2  -	Out Low		User LED 0
#define B_SDD1	 		BIT3		// PB3  B	-	-		SD Card Data 1 
#define B_SDD2	 		BIT4		// PB4  B	-	-		SD Card Data 2
#define B_SDD3	 		BIT5		// PB5  B	-	-		SD Card Data 3
#define PB6	 			38			// PB6  - 	In	-		Expansion GPIO
#define PB7	 			39			// PB7  - 	In	-		Expansion GPIO
#define PB8	 			40			// PB8  - 	In	-		Expansion GPIO
#define PB9	 			41			// PB9  - 	In	-		Expansion GPIO
#define PB10	 		42			// PB10 - 	In	-		Expansion GPIO
#define PB11	 		43			// PB11 - 	In	-		Expansion GPIO
#define B_ETXD2	 		BIT12		// PB12 B	-	-		Ethernet Transmit Data 2
#define B_ETXD3	 		BIT13		// PB13 B	-	-		Ethernet Transmit Data 3
#define B_ETXER	 		BIT14		// PB14 B	-	-		Ethernet Transmit Error
#define B_ERXD2	 		BIT15		// PB15 B	-	-		Ethernet Receive Data  
#define B_ERXD3	 		BIT16		// PB16 B	-	-		Ethernet Receive Data 
#define B_ERXDV	 		BIT17		// PB17 B	-	-		Ethernet Receive Data Valid
#define B_ECOL	 		BIT18		// PB18 B	-	-		Ethernet Collision
#define B_ERXCK	 		BIT19		// PB19 B	-	-		Ethernet Receive Clock
#define A_U1TX	 		BIT20		// PB20 A	-	-		USART1 Transmit
#define A_U1RX	 		BIT21		// PB21 A	-	-		USART1 Receive
#define PB22	 		54			// PB22 -	In	-		Expansion GPIO
#define A_U1DCD	 		BIT23		// PB23 A	-	-		USART1 DCD 
#define A_U1CTS	 		BIT24		// PB24 A	-	-		USART1 CTS
#define A_U1DSR	 		BIT25		// PB25 A	-	-		USART1 DSR
#define A_U1RTS	 		BIT26		// PB26 A	-	-		USART1 RTS
#define A_PCK0			BIT27		// PB27 A	-	-		Peripheral Clock 0 - Bus Clock
#define USR_SW1			60			// PB28 -	In	Low		User Switch 1
#define USR_SW0			61			// PB29 -	In	Low		User Switch 0
// PB30 - Does not exist							
// PB31 - Does not exist							

// default settings for PORT B
// Peripheral A select - USART 1, Peripheral Clock 0
#define DEF_PIOB_ASR	(A_U1TX | A_U1RX | A_U1DCD | A_U1CTS | A_U1DSR | A_U1RTS | A_PCK0)

// Peripheral B select - SD Card 0 and Ethernet
#define DEF_PIOB_BSR	(B_SDD1 | B_SDD2 | B_SDD3 | B_ETXD2 | B_ETXD3 | B_ETXER | B_ERXD2 | B_ERXD3 | B_ERXDV | B_ECOL | B_ERXCK)

 // enable pullups where desired - usually just expansion GPIO
#define DEF_PIOB_PUER	(PB6 | PB7 | PB8 | PB9 | PB10 | PB11 | PB22)

// enable default state of outputs that need to start high
#define DEF_PIOB_SODR	(USR_LED2 | USR_LED1 | USR_LED0)

// enable default state of outputs that need to start low
#define DEF_PIOB_CODR	0

// enable all pins that start as outputs
#define DEF_PIOB_OER	(USR_LED2 | USR_LED1 | USR_LED0)

// PORT C							   BIT	A/B	DIR	SENSE	DESCRIPTION
#define XINT0	 		64			// PC0	-	In	Low		Expansion Interrupt 0
#define XINT1	 		65			// PC1	-	In	Low		Expansion Interrupt 1
#define PHYINT	 		66			// PC2	-	In	Low		LT971A PHY Interrupt
#define PC3	 			67			// PC3	-	-	-		Unused GPIO
#define CINT0	 		68			// PC4	-	In	Low		CAN Interrupt 0
#define CINT1	 		69			// PC5	-	In	Low		CAN Interrupt 1
#define A_WAIT	 		BIT6		// PC6	A	-	-		WAIT Input	
#define A_A23	 		BIT7		// PC7	A	-	-		Address Bus 23
#define PC8	 			72			// PC8  - 	In	-		Expansion GPIO
#define PC9	 			73			// PC9  - 	In	-		Expansion GPIO
#define PC10	 		74			// PC10 - 	In	-		Expansion GPIO
#define PC11	 		75			// PC11 - 	In	-		Expansion GPIO
#define PC12	 		76			// PC12 - 	In	-		Expansion GPIO
#define A_CS7	 		BIT13		// PC13	-	-	-		Chip Select 7 (used as Expansion Chip Select 1)
#define RI1		 		78			// PC14	-	In	Low		USART 1 RI - GPIO mode
#define DTR1	 		79			// PC15	-	In	Low		USART 1 DTR - GPIO mode
#define A_D16	 		BIT16		// PC16	A	-	-		Data Bus 16
#define A_D17	 		BIT17		// PC17	A	-	-		Data Bus 17
#define A_D18	 		BIT18		// PC18	A	-	-		Data Bus 18
#define A_D19	 		BIT19		// PC19	A	-	-		Data Bus 19
#define A_D20	 		BIT20		// PC20	A	-	-		Data Bus 20
#define A_D21	 		BIT21		// PC21	A	-	-		Data Bus 21
#define A_D22	 		BIT22		// PC22	A	-	-		Data Bus 22
#define A_D23	 		BIT23		// PC23	A	-	-		Data Bus 23
#define A_D24	 		BIT24		// PC24	A	-	-		Data Bus 24
#define A_D25	 		BIT25		// PC25	A	-	-		Data Bus 25
#define A_D26	 		BIT26		// PC26	A	-	-		Data Bus 26
#define A_D27	 		BIT27		// PC27	A	-	-		Data Bus 27
#define A_D28	 		BIT28		// PC28	A	-	-		Data Bus 28
#define A_D29	 		BIT29		// PC29	A	-	-		Data Bus 29
#define A_D30	 		BIT30		// PC30	A	-	-		Data Bus 30
#define A_D31	 		BIT31 		// PC31	A	-	-		Data Bus 31
// default settings for PORT C
// Peripheral A select - Wait, A23 and Databus
#define DEF_PIOC_ASR	(A_WAIT | A_A23 | 0xffff0000)

// Peripheral B select - None
#define DEF_PIOC_BSR	0

 // enable pullups where desired - usually just expansion GPIO
#define DEF_PIOC_PUER	(PC3 | PC8 | PC9 | PC10 | PC11 | PC12)

// enable default state of outputs that need to start high
#define DEF_PIOC_SODR	0	// no outputs assigned in PortC

// enable default state of outputs that need to start low
#define DEF_PIOC_CODR	0	// no outputs assigned in PortC

// enable all pins that start as outputs
#define DEF_PIOC_OER	0	// no outputs assigned in PortC

// PORT D					   	   BIT	A/B	DIR	SENSE	DESCRIPTION
#define PD0	 			96			// PD0	-	In	-		Expansion GPIO
#define PD1	 			97			// PD1	-	In	-		Expansion GPIO
#define PD2	 			98			// PD2	-	In	-		Expansion GPIO
#define PD3	 			99			// PD3	-	In	-		Unused GPIO
#define PD4	 			100			// PD4	-	In	-		Unused GPIO
#define SD_CD			101			// PD5	-	In	Low		SD Card Detect
#define SD_WP			102			// PD6	-	In	Low		SD Card Write Protect
#define B_ETMSYNC		BIT7		// PD7	B	-	-		ETM Sync 
#define B_ETMCLK		BIT8		// PD8	B	-	-		ETM Clock
#define B_ETMPS0		BIT9		// PD9	B	-	-		ETM Processor Status Bit 0
#define B_ETMPS1		BIT10		// PD10	B	-	-		ETM Processor Status Bit 1
#define B_ETMPS2		BIT11		// PD11	B	-	-		ETM Processor Status Bit 2
#define B_ETMPK0		BIT12		// PD12	B	-	-		ETM Packet Bit 0
#define B_ETMPK1		BIT13		// PD13	B	-	-		ETM Packet Bit 1
#define B_ETMPK2		BIT14		// PD14	B	-	-		ETM Packet Bit 2
#define B_ETMPK3		BIT15		// PD15	B	-	-		ETM Packet Bit 3
#define B_ETMPK4		BIT16		// PD16	B	-	-		ETM Packet Bit 4
#define B_ETMPK5		BIT17		// PD17	B	-	-		ETM Packet Bit 5
#define B_ETMPK6		BIT18		// PD18	B	-	-		ETM Packet Bit 6
#define B_ETMPK7		BIT19		// PD19	B	-	-		ETM Packet Bit 7
#define B_ETMPK8		BIT20		// PD20	B	-	-		ETM Packet Bit 8
#define B_ETMPK9		BIT21		// PD21	B	-	-		ETM Packet Bit 9
#define B_ETMPK10		BIT22		// PD22	B	-	-		ETM Packet Bit 10
#define B_ETMPK11		BIT23		// PD23	B	-	-		ETM Packet Bit 11
#define B_ETMPK12		BIT24		// PD24	B	-	-		ETM Packet Bit 12
#define B_ETMPK13		BIT25		// PD25	B	-	-		ETM Packet Bit 13
#define B_ETMPK14		BIT26		// PD26	B	-	-		ETM Packet Bit 14
#define B_ETMPK15		BIT27		// PD27	B	-	-		ETM Packet Bit 15
// PD28 Does Not Exist
// PD29 Does Not Exist
// PD30 Does Not Exist
// PD31 Does Not Exist

// default settings for PORT D
// Peripheral A select - none
#define DEF_PIOD_ASR	0

// Peripheral B select - Embedded Trace Module
// We probably shouldn't enable ETM by default as it can be turned on from JTAG.
#define DEF_PIOD_BSR	0

 // enable pullups on all but the two CD Card status bits
#define DEF_PIOD_PUER	~(SD_CD | SD_WP)

// enable default state of outputs that need to start high
#define DEF_PIOD_SODR	0

// enable default state of outputs that need to start low
#define DEF_PIOD_CODR	0

// enable all pins that start as outputs
#define DEF_PIOD_OER	0


#endif // #ifndef CSB337_GPIO_H

