/* cpuio.c:
 * This file contains the generic serial port drivers and other
 * miscellaneous target specific functions.
 *
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 *
 * Author:  Ed Sutter
 * email:   esutter@lucent.com
 * phone:   908-582-2351
 */
#include "config.h"
#include "genlib.h"
#include "cpuio.h"
#include "stddefs.h"
#include "ppcuic.h"
#include "xio.h"
#include "uart16550.h"

int
devInit(int baud)
{
	extern int i2c_init(void);

	/* If the console is configured for 'B', then we need to steer
	 * it to RS232...
	 */
	if (CONSOLE_UART_BASE == UART_B)
		*(ulong *)0xff000008 |= 0xe0000000;	

	// Initialize the I2C Port
	i2c_init();

	// Initialize the UART
    InitUART(baud);

	return 0;
}

void
initCPUio()
{
	ppcMtuicsr(0xFFFFFFFF);		/* clear all ints */
	ppcMtuicer(0x00000000);		/* disable all ints */
	ppcMtuiccr(0);				/* Set all ints to non-critical. */
	ppcMtuicpr(0xFFFFE07F);		/* Set up positive interrupt polarity. */
	ppcMtuicsr(0xFFFFFFFF);		/* clear all ints */

	xio_init();
}

/* uSleep():
 * Reasonably accurate microsecond-resolution delay function hooked to
 * the Low Time Base register of the PPC405.  It assumes the time base
 * is derived from the CPU clock, so CPC0_CR1[CETE] should be low.
 * Example usage of GNU-only syntax for using a C-variable in a
 * line of assembler:
 *  read:
 *  asm volatile ("mfdcr %0, 0xa0" : "=r" (tmp)); 
 *  write:
 *  asm volatile ("mtdcr 0xa0, %0" : "=r" (tmp) : "0" (tmp));
 */
void
uSleep(int delay)
{
	unsigned duration;
	unsigned register begin, end;

	duration = delay * (MON_CPU_CLOCK/1000000);

	/* Retrieve value from TBL first time...
	 */
	asm volatile ("mfspr %0, 0x10c" : "=r" (begin)); /* TBL */

	/* Poll TBL until the delta is 1/1000th of a second...
	 */
	do {
		asm volatile ("mfspr %0, 0x10c" : "=r" (end)); /* TBL */
	} while((end - begin) < duration);
}

void
csb472_setup(void)
{
#if FOR_COGENT
	extern void watchdog_check(void);

	watchdog_check();
#endif

#ifdef FLASH_IS_AMD
	setenv("FLASH_TYPE","AMD");
#else
	setenv("FLASH_TYPE","INTEL");
#endif

}
