/* etherdev.c:
 * Ethernet driver for CSB655...
 * This driver is a re-write from the original that I got from the
 * Alchemy evaluation board's monitor.  I beleive it follows closely
 * to the recommendation outlined in the Alchemy Document titled:
 * AMD Alchemy (tm) AU1550 Processor Databook" dated September 2004,
 * with publication ID 30283B.
 * As of June 4, 2005, the driver is working pretty well; however, I
 * don't have a handle on the PHY at this point.  I think it is just
 * coming up and running on its own.  I can query the ID and I get the
 * expected values; however, the status register makes no sense.  It
 * indicates that both 10 & 100 are active and HD & FD as well.
 *
 *	General notice:
 *	This code is part of a boot-monitor package developed as a generic base
 *	platform for embedded system designs.  As such, it is likely to be
 *	distributed to various projects beyond the control of the original
 *	author.  Please notify the author of any enhancements made or bugs found
 *	so that all may benefit from the changes.  In addition, notification back
 *	to the author will allow the new user to pick up changes that may have
 *	been made by other users after this version of the code was distributed.
 *
 *	Note1: the majority of this code was edited with 4-space tabs.
 *	Note2: as more and more contributions are accepted, the term "author"
 *		   is becoming a mis-representation of credit.
 *
 *	Original author:	Ed Sutter
 *	Email:				esutter@lucent.com
 *	Phone:				908-582-2351
 */
#include "config.h"
#if INCLUDE_ETHERNET
#include "stddefs.h"
#include "genlib.h"
#include "ether.h"
#include "phy.h"
#include "mips.h"

/* There are two MACs in the AU1550, each has its own set of base
 * address registers...
 */
#define MAC0_BASE			0xB0500000 
#define MAC0_EN_BASE		0xB0520000
#define MAC0_DMA_BASE		0xB4004000

#define MAC1_BASE			0xB0510000
#define MAC1_EN_BASE		0xB0520004
#define MAC1_DMA_BASE		0xB4004200

/* These additional ADDRLO/ADDRHI definitions are used at startup
 * so that both MACs are loaded with an ethernet address.
 * Even though uMon only uses one (or the other) MAC, we still load
 * both MACs with an ethernet address.
 */
#define MAC0_ADDRLO	(MAC0_BASE+8)
#define MAC0_ADDRHI	(MAC0_BASE+4)
#define MAC1_ADDRLO	(MAC1_BASE+8)
#define MAC1_ADDRHI	(MAC1_BASE+4)

/* MII Control bits:
 */
#define MII_BUSY			0x00000001
#define MII_WRITE			0x00000002

/* MAC Enable register bits:
 */
#define MAC_ENABLE_DMARESET	0x00000040	// DMA (ae_b) Reset
#define MAC_ENABLE_TXEN		0x00000020	// Value is anded with TX_EN
#define MAC_ENABLE_MIIENAB	0x00000010	// 0-Assert 1-Deassert Rst_T/RxClkN
#define MAC_ENABLE_MACINIT	0x00000002	// 0-> Resets MAC, 1-> Enables MAC
#define MAC_ENABLE_MACCLK	0x00000001	// 1-> Enables MAC clock

/* MAC Control register bits:
 */
#define MAC_CONTROL_PR 		0x00040000	// Promiscuous
#define MAC_CONTROL_EM 		0x40000000	// Endian
#define MAC_CONTROL_DO 		0x00800000	// Disable receive own
#define MAC_CONTROL_FD 		0x00100000	// Full duplex 
#define MAC_CONTROL_DB	 	0x00000800	// Broadcast
#define MAC_CONTROL_TE	 	0x00000008	// Transmit enable
#define MAC_CONTROL_RE	 	0x00000004	// Receive enable

/* Ethernet receive status bits:
 */
#define MISSED_FRAME		0x80000000
#define PACKET_FILTER		0x40000000
#define TRANSACTION_DONE	0x00000002
#define DMA_ENABLE			0x00000001

/* Miscellaneous:
 */
#define MIN_PKTSIZE			60
#define MAX_PKTSIZE			2048
#define MEMBLOCKSIZE		(MAX_PKTSIZE * 8)

struct mac_rxdma {
	vulong *stat;
	vulong *addr;
};

struct mac_txdma {
	vulong *stat;
	vulong *addr;
	vulong *len;
};

struct au1550_enet {
	vulong *ctrl;
	vulong *addrhi;
	vulong *addrlo;
	vulong *hashhi;
	vulong *hashlo;
	vulong *miictrl;
	vulong *miidata;
	vulong *flowctrl;
	vulong *vlan1;
	vulong *vlan2;
	vulong *enable;
	struct mac_txdma txdma[4];
	struct mac_rxdma rxdma[4];
} edev;

struct edev_stat {
	int multicast;
	int collisions;
	int rx_len_err;
	int rx_cr_err;
	int rx_db_err;
	int rx_fl_err;
	int tx_fa_err;
	int tx_ls_err;
	int tx_ur_err;
	int tx_lc_err;
	int tx_ed_err;
} estat;

static int		phy_addr;		// address of phy on MII
static int		phy_num;		// 0 or 1
static int		rIdx, tIdx;		// buffer idx
static vulong	xmtBuf[4];		// transmit buffers
static vulong	rcvBuf[4];		// receive buffers
static unsigned char memblock[MEMBLOCKSIZE+32];

/* Non-cached memory is still subject to the write buffer.
 * To guarantee that a write to a MAC register has been
 * posted, you must flush the write buffer using the sync
 * instruction.  see AMD Au1550 Data Book, section 9.4.3.1
 */
void static inline au_sync(void)
{
	__asm__ volatile ("sync");
}

/* phy_read() & phy_write():
 * Talk to the PHY through the MII interface.
 * For each access, wait for BUSY to clear.
 */
int
phy_read(int phy, int reg, ushort *value)
{
	int giveup;

	/* Wait till PHY-not-busy...
	 */
	giveup = 100;
	while(*edev.miictrl & MII_BUSY) {
		monDelay(1);
		if (--giveup <= 0) {
			printf("phy_read() busy wait1 fail\n");
			return(-1);
		}
	}

	*(edev.miictrl) = (((phy_addr&0x1f) << 11) | ((reg&0x1f) << 6));
	au_sync();

	/* check link status, spin here till done */
	giveup = 100;
	while(*edev.miictrl & MII_BUSY) {
		monDelay(1);
		if (--giveup <= 0) {
			printf("phy_read() busy wait2 fail\n");
			return(-1);
		}
	}
	*value = *edev.miidata;
	return(0);
}

int
phy_write(int phy, int reg,ushort val)
{
	int giveup;

	/* Wait till PHY-not-busy...
	 */
	giveup = 1000;
	while(*edev.miictrl & MII_BUSY) {
		monDelay(1);
		if (--giveup <= 0) {
			printf("phy_write() busy wait1 fail\n");
			return(-1);
		}
	}

	*edev.miidata = (ulong)val;
	*edev.miictrl = (((phy_addr&0x1f) << 11) | ((reg&0x1f) << 6) | MII_WRITE);
	au_sync();

	/* check link status, spin here till done */
	giveup = 1000;
	while(*edev.miictrl & MII_BUSY) {
		monDelay(1);
		if (--giveup <= 0) {
			printf("phy_write() busy wait2 fail\n");
			return(-1);
		}
	}
	return(0);
}

static int
macAllocBuf(void)
{
	int	i;
	void	*bp;

	/* All memory buffers must be implemented on a cache-line boundary
	 * (32 bytes)...
	 */
	bp = memblock;
	while((int)bp & 0x1f)
		bp++;

	bp = (void*) KSEG1((ulong)bp);
	memset(bp,0,MEMBLOCKSIZE);

	/* Init RX buffer pointers
	 */
	for (i=0;i<4;i++) {
		rcvBuf[i] = PHYS((ulong) bp);
		bp += MAX_PKTSIZE;
	}

	/* Init TX buffer pointers
	 */
	for (i=0;i<4;i++) {
		xmtBuf[i] = PHYS((ulong) bp);
		bp += MAX_PKTSIZE;
	}

	return(0);
}

void
macShowStatus(void)
{
	if (ENET_DEBUG_ENABLED())
		phy_dump(phy_num);

	phy_stat(phy_num,1);

	printf(" Multicasts received:            %ld\n\r", estat.multicast);
	printf(" Collisions:                     %ld\n\r", estat.collisions);
	printf(" Receive length errors:          %ld\n\r", estat.rx_len_err);
	printf(" Receive CRC errors:             %ld\n\r", estat.rx_cr_err);
	printf(" Receive Dribble errors:         %ld\n\r", estat.rx_db_err);
	printf(" Receive frame length:		     %ld\n\r", estat.rx_fl_err);
	printf(" Transmit aborted errors:        %ld\n\r", estat.tx_fa_err);
	printf(" Transmit lost carrier errors:   %ld\n\r", estat.tx_ls_err);
	printf(" Transmit FIFO underrun errors:  %ld\n\r", estat.tx_ur_err);
	printf(" Transmit late collision errors:  %ld\n\r", estat.tx_lc_err);
	printf(" Transmit timeout errors:        %ld\n\r", estat.tx_ed_err);
};

void
ShowEtherdevStats(void)
{
	macShowStatus();
}

void
edev_init(ulong ctrl_base, ulong enable_base, ulong dma_base)
{
	int i;

	/* Initialize the entire enet data structure based on the above
	 * base address settings:
	 */
	edev.ctrl = (vulong *)ctrl_base;				// MAC control
	edev.addrhi = (vulong *)(ctrl_base + 0x0004);	// MAC address high
	edev.addrlo = (vulong *)(ctrl_base + 0x0008);	// MAC address low
	edev.hashhi = (vulong *)(ctrl_base + 0x000c);	// Multicast HiAddr
	edev.hashlo = (vulong *)(ctrl_base + 0x0010);	// Multicast LoAddr
	edev.miictrl = (vulong *)(MAC0_BASE + 0x0014);	// MII address
	edev.miidata = (vulong *)(MAC0_BASE + 0x0018);	// MII data
	edev.flowctrl = (vulong *)(ctrl_base + 0x001c);	// Flow control
	edev.vlan1 = (vulong *)(ctrl_base + 0x0020);	// VLAN1 tag
	edev.vlan2 = (vulong *)(ctrl_base + 0x0024);	// VLAN2 tag
	edev.enable = (vulong *)enable_base;
	for(i=0;i<4;i++) {
		edev.txdma[i].stat = (vulong *)(dma_base + (i*0x10));
		edev.txdma[i].addr = (vulong *)(dma_base + (i*0x10) + 4);
		edev.txdma[i].len = (vulong *)(dma_base + (i*0x10) + 8);
	}
	for(i=0;i<4;i++) {
		edev.rxdma[i].stat = (vulong *)(dma_base + 0x100 + (i*0x10));
		edev.rxdma[i].addr = (vulong *)(dma_base + 0x100 + (i*0x10) + 4);
	}
}

void
edev_print(void)
{
	int i;

	printf("phy_addr: %d\n",phy_addr);
	printf("ctrl:     0x%lx\n",(long)edev.ctrl);
	printf("addrhi:   0x%lx\n",(long)edev.addrhi);
	printf("addrlo:   0x%lx\n",(long)edev.addrlo);
	printf("hashhi:   0x%lx\n",(long)edev.hashhi);
	printf("hashlo:   0x%lx\n",(long)edev.hashlo);
	printf("miictrl:  0x%lx\n",(long)edev.miictrl);
	printf("miidata:  0x%lx\n",(long)edev.miidata);
	printf("flowctrl: 0x%lx\n",(long)edev.flowctrl);
	printf("vlan1:    0x%lx\n",(long)edev.vlan1);
	printf("vlan2:    0x%lx\n",(long)edev.vlan2);
	printf("enable:   0x%lx\n",(long)edev.enable);
	for(i=0;i<4;i++) {
		printf("txdma[%d].stat 0x%lx\n",i,(long)edev.txdma[i].stat);
		printf("txdma[%d].addr 0x%lx\n",i,(long)edev.txdma[i].addr);
		printf("txdma[%d].len  0x%lx\n",i,(long)edev.txdma[i].len);
	}
	for(i=0;i<4;i++) {
		printf("rxdma[%d].stat 0x%lx\n",i,(long)edev.rxdma[i].stat);
		printf("rxdma[%d].addr 0x%lx\n",i,(long)edev.rxdma[i].addr);
	}
}

void
macDmaStop(void)
{
	int i;

	/* Disable all DMA:
	 */
	for(i=0;i<4;i++) 
		*(edev.txdma[i].addr) &= ~DMA_ENABLE;
	
	for(i=0;i<4;i++) 
		*(edev.rxdma[i].addr) &= ~DMA_ENABLE;

	au_sync();
}

void
macDmaInit(void)
{
	int i;

	/* MAC DMA Init:
	 */
	for (i=0;i<4;i++)
		*edev.rxdma[i].addr = rcvBuf[i] | DMA_ENABLE;

	for (i=0;i<4;i++) {
		*edev.txdma[i].addr = 0;
		*edev.txdma[i].len = 0;
	}

	au_sync();
}

/*
 * enreset():
 *	Reset the PHY and MAC.
 */
void
enreset(void)
{
	ulong ctrl_base, enable_base, dma_base;

	/* Start by disabling both MACs...
	 */
	*(vulong *)MAC0_EN_BASE = MAC_ENABLE_DMARESET;
	*(vulong *)MAC1_EN_BASE = MAC_ENABLE_DMARESET;
	au_sync();

	/* Establish connectivity with one of the two MACs...
	 * Note that MAC0 is labeled ENET1 on early spins of the CSB655's
	 * carrier board.  MAC1 is the default, its the RJ45 closest to the
	 * Compact Flash connector.
	 */
#ifdef ORIG_CSB655
	if (getenv("USE_MAC0")) {
		phy_addr = 4;
		phy_num = 0;
		ctrl_base= (ulong)MAC0_BASE;
		enable_base= (ulong)MAC0_EN_BASE;
		dma_base= (ulong)MAC0_DMA_BASE;
	}
	else {
		phy_addr = 3;
		phy_num = 1;
		ctrl_base= (ulong)MAC1_BASE;
		enable_base= (ulong)MAC1_EN_BASE;
		dma_base= (ulong)MAC1_DMA_BASE;
	}
#else
	if (getenv("USE_MAC1")) {
		phy_addr = 1;
		phy_num = 1;
		ctrl_base= (ulong)MAC1_BASE;
		enable_base= (ulong)MAC1_EN_BASE;
		dma_base= (ulong)MAC1_DMA_BASE;
	}
	else {
		phy_addr = 0;
		phy_num = 0;
		ctrl_base= (ulong)MAC0_BASE;
		enable_base= (ulong)MAC0_EN_BASE;
		dma_base= (ulong)MAC0_DMA_BASE;
	}

#endif

	edev_init(ctrl_base, enable_base, dma_base);
	if (ENET_DEBUG_ENABLED())
		edev_print();

	macDmaStop();

	/* Enable clocks and pull MAC out of reset...
	 * This has to be done otherwise access to the memory mapped space 
	 * will lock up the CPU.
	 */
	*(vulong *)MAC0_EN_BASE = (MAC_ENABLE_MACCLK | MAC_ENABLE_DMARESET);
	*(vulong *)MAC1_EN_BASE = (MAC_ENABLE_MACCLK | MAC_ENABLE_DMARESET);
	au_sync();
	monDelay(10);
	*(vulong *)MAC0_EN_BASE= (MAC_ENABLE_TXEN | MAC_ENABLE_MIIENAB | \
				   MAC_ENABLE_MACINIT | MAC_ENABLE_MACCLK);
	*(vulong *)MAC1_EN_BASE= (MAC_ENABLE_TXEN | MAC_ENABLE_MIIENAB | \
				   MAC_ENABLE_MACINIT | MAC_ENABLE_MACCLK);
	au_sync();
	monDelay(50);
//	phy_write(phy_num,PHY_CTRL,PHY_CTRL_RST);
//	monDelay(100);
}

int
eninit(void)
{
	uchar	byte;
	ushort id1, id2;

	/* Clear the driver-specific error counts:
	 */
	memset((char *)&estat,0,sizeof(struct edev_stat));

	*(edev.ctrl) =  MAC_CONTROL_PR | MAC_CONTROL_DO | MAC_CONTROL_FD;
#ifdef EB
	*(edev.ctrl) |=  MAC_CONTROL_EM;
#endif
	au_sync();

	/* Load MAC address:
	 * The CSB655 has two ethernet ports.  To support allocation of two
	 * MAC addresses on this board, we use the assigned ETHERADD value
	 * for MAC0 and ETHERADD+1 for MAC1...
	 */
	*(vulong *)MAC0_ADDRLO = BinEnetAddr[0] | (BinEnetAddr[1] << 8)   |
		(BinEnetAddr[2] << 16)  | (BinEnetAddr[3] << 24);
	*(vulong *)MAC0_ADDRHI = BinEnetAddr[4] | (BinEnetAddr[5] << 8);
	*(vulong *)MAC1_ADDRLO = BinEnetAddr[0] | (BinEnetAddr[1] << 8)   |
		(BinEnetAddr[2] << 16)  | (BinEnetAddr[3] << 24);
	byte = BinEnetAddr[5] + 1;
	*(vulong *)MAC1_ADDRHI = BinEnetAddr[4] | (byte << 8);
	au_sync();

	macAllocBuf();
	macDmaInit();

	/* Read PHY ID to verify connectivity...	
 	 */
	phy_read(phy_num,PHY_IDHI,&id1);
	phy_read(phy_num,PHY_IDLO,&id2);
	if ((id1 != 0x0040) || (id2 != 0x6322))
		printf("WARNING: Unexpected phy_id: %04x %04x\n",id1,id2);

	*edev.ctrl |= MAC_CONTROL_TE | MAC_CONTROL_RE; 
	au_sync();

	/* The initial state of the current buffer (CB) index is
	 * unknown, thus must be read from the device.
	 */
	rIdx = (*edev.rxdma[0].addr & 0xC) >> 2;
	tIdx = (*edev.txdma[0].addr & 0xC) >> 2;

	return(0) ;
}

int
EtherdevStartup(int verbose)
{
	enreset();
	return(eninit());
}

/* getXmitBuffer():
 * Return a pointer to the buffer that is to be used for transmission of
 * the next packet.  Since the monitor's driver is EXTREMELY basic,
 * there will only be one packet ever being transmitted.  No need to queue
 * up transmit packets.
 */
uchar *
getXmitBuffer(void)
{
	return((uchar *)KSEG1(xmtBuf[tIdx]));
}

/* sendBuffer():
 * Send out the packet assumed to be built in the buffer returned by the
 * previous call to getXmitBuffer() above.
 */
int
sendBuffer(int pktlen)
{
	int		i;
	char	*packet;
	static	int lastidx;

	packet = (char*)KSEG1(xmtBuf[tIdx]);

#if INCLUDE_ETHERVERBOSE
	if (EtherVerbose &  SHOW_OUTGOING)
		printPkt((struct ether_header *)packet,pktlen,ETHER_OUTGOING);
#endif

	/* If a previous packet was sent, then make sure it's done
	 * before sending another packet...
	 */
	if (EtherXFRAMECnt > 0) {
		for(i=0;i<1000;i++) {
			if (*edev.txdma[lastidx].addr & TRANSACTION_DONE)
				break ;
			monDelay(1);
		}
		if (i >= 1000) {
			if (ENET_DEBUG_ENABLED())
				printf(" MAC TX Timeout - Re-init!\n") ;
			//enreset();
			//eninit();
			//return(-1) ;
		}

		/* Clear status and done bit */
		*edev.txdma[lastidx].stat = 0;
		*edev.txdma[lastidx].addr &= ~TRANSACTION_DONE;
		au_sync();
	}

	/* If either DONE or ENABLE are not 0 then DMA channel is not ready
	 * tIdx picks correct DMA register...
	 */
	if (*edev.txdma[tIdx].addr & (DMA_ENABLE | TRANSACTION_DONE)) {
		if (ENET_DEBUG_ENABLED())
			printf("MAC TX DMA not ready!\n");
		return(-1);
	}

	EtherXFRAMECnt++;
 
	/* check to PAD packet up to minimum size
	 */
	if (pktlen < MIN_PKTSIZE) {
		memset((packet + pktlen),0,(MIN_PKTSIZE-pktlen) ) ;
		pktlen = MIN_PKTSIZE ;
	}

	/* Set length register and enable the packet transfer...
	 */
	*edev.txdma[tIdx].len = pktlen;
	*edev.txdma[tIdx].addr = xmtBuf[tIdx] | DMA_ENABLE;
	au_sync();

	lastidx = tIdx;
	tIdx++;
	if (tIdx == 4)
		tIdx = 0;
	return(0) ;
}

/*
 * polletherdev():
 * Called continuously by the monitor (ethernet.c) to determine if there
 * is any incoming ethernet packets.
 */
int
polletherdev(void)
{
	int	idx, len, pktcnt = 0;
	ulong	status;
	struct ether_header	*epkt;

	idx = rIdx;
	if (*edev.rxdma[idx].addr & TRANSACTION_DONE) {
		status = *edev.rxdma[idx].stat;
		if (ENET_DEBUG_ENABLED())
			printf("rcv status (%d): 0x%lx\n",idx,status);
		if (status & MISSED_FRAME) {
			rIdx++;
			if (rIdx == 4)
				rIdx = 0;
		}
		else if (status & PACKET_FILTER) {
			epkt = (struct ether_header *)KSEG0(rcvBuf[idx]);
			len = status & 0x3fff;
			EtherRFRAMECnt++;
			pktcnt++;
			rIdx++;
			if (rIdx == 4)
				rIdx = 0;
			processPACKET(epkt,len);
		}
		else {
			target_putchar('?');
		}
		*edev.rxdma[idx].addr = rcvBuf[idx] | DMA_ENABLE;
		au_sync();
	}
	return(pktcnt);
}

void
disablePromiscuousReception(void)
{
	*edev.ctrl &= ~MAC_CONTROL_PR;
	au_sync();
}

void
enablePromiscuousReception(void)
{
	*edev.ctrl |= MAC_CONTROL_PR;
	au_sync();
}

void
disableBroadcastReception(void)
{
	*edev.ctrl |= MAC_CONTROL_DB;
	au_sync();
}

void
enableBroadcastReception(void)
{
	*edev.ctrl &= ~MAC_CONTROL_DB;
	au_sync();
}

/* 
 * enselftest():
 *	Run a self test of the ethernet device(s).  This can be stubbed
 *	with a return(1).
 *	Return 1 if success; else -1 if failure.
 */
int
enselftest(int verbose)
{
	printf("Self-test unavailable for this device\n");
	return(0);
}

/* DisableEtherdev():
 * Fine as it is...
 */
void
DisableEtherdev(void)
{
	/* Disable both MACs...
	 */
	*(vulong *)MAC0_EN_BASE = 0x00000040;
	*(vulong *)MAC1_EN_BASE = 0x00000040;
	au_sync();
}

/* extGetEtherAdd():
 * If there was some external mechanism (other than just using the
 * ETHERADD shell variable established in the monrc file) for retrieval of
 * the board's MAC address, then do it here...
 */
char *
extGetEtherAdd(void)
{
	return((char *)0);
}

char *
extGetIpAdd(void)
{
	return((char *)0);
}

#endif
