//==========================================================================
//
//      cpu_ipu.c
//
// Author(s):   ORIGINAL Code from Freescale LINUX, modified by Yossi Har-Nov
// Date:        03/28/2007
// Description:	Image Processing Unit (IPU) routines for i.MX31
//
//==========================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "stddefs.h"
#include "cli.h"
#include "cpu_ipu.h"
#include "mx31_lcd.h"

#define NULL 0x0000

/* Globals */
ulong         gOpenCount;        // Counter to track opens.
ulong        g_ipu_clk;
ulong        g_ipu_csi_clk;
int          gSecChanEn[21];
ulong        gChannelInitMask;

static ulong g_hStartWidth;
static ulong g_vStartWidth;

static const ulong DI_mappings[] =
{
	0x1600AAAA, 0x00E05555, 0x00070000, 3,	// RGB888
	0x0005000F, 0x000B000F, 0x0011000F, 1,	// RGB666
#if 1
	0x0004003F, 0x000A000F, 0x000F003F, 1	// RGB565
#endif
//	0x0005003F, 0x000B000F, 0x0011003F, 1	// RGB565
};

static ulong channel_2_dma(ipu_channel_t ch, ipu_buffer_t type)
{
    return ((type==IPU_INPUT_BUFFER) ? ((ulong)ch & 0xFF) : \
           ((type==IPU_OUTPUT_BUFFER) ? (((ulong)ch>>8) & 0xFF) : (((ulong)ch>>16) & 0xFF)));
}

static ulong DMAParamAddr(ulong dma_ch)
{
    return ( 0x10000 | (dma_ch << 4) );
}

/*!
 * This function is called to initialize a logical IPU channel.
 * 
 * @param       channel Input parameter for the logical channel ID to initalize.
 *
 * @param       params  Input parameter containing union of channel initialization 
 *                      parameters.
 *
 * @return      This function returns 0 on success or negative error code on fail
 */
ulong ipu_init_channel(ipu_channel_t channel, ipu_channel_params_t * params )
{
    ulong ipu_conf;
    //ulong reg;
    //ulong lock_flags;

#if(0)
    if ((channel != MEM_SDC_BG) && (channel != MEM_SDC_FG) && 
        (channel != MEM_ROT_ENC_MEM) && (channel != MEM_ROT_VF_MEM) &&
        (channel != MEM_ROT_PP_MEM) && (channel != CSI_MEM) 
        && (params == NULL)) {
        return -1;
    }
#endif

    gChannelInitMask |= 1L << IPU_CHAN_ID(channel);
    
    // Enable IPU sub module
    ipu_conf = IPU_REG(IPU_CONF);
    if (gChannelInitMask & 0x00000066L) //CSI
    {
        ipu_conf |= IPU_CONF_CSI_EN;
    }
    if (gChannelInitMask & 0x00001FFFL) //IC
    {
        ipu_conf |= IPU_CONF_IC_EN;
    }
    if (gChannelInitMask & 0x00000A10L) //ROT
    {
        ipu_conf |= IPU_CONF_ROT_EN;
    }
    if (gChannelInitMask & 0x0001C000L) //SDC
    {
        ipu_conf |= IPU_CONF_SDC_EN | IPU_CONF_DI_EN;
    }
    if (gChannelInitMask & 0x00061140L) //ADC
    {
        ipu_conf |= IPU_CONF_ADC_EN | IPU_CONF_DI_EN;
    }
    if (gChannelInitMask & 0x00380000L) //PF
    {
        ipu_conf |= IPU_CONF_PF_EN;
    }
	IPU_REG(IPU_CONF) = ipu_conf;
    
    return 0;
}

ulong bytes_per_pixel(ulong fmt)
{
    switch (fmt)
    {
        case IPU_PIX_FMT_GENERIC: //generic data
        case IPU_PIX_FMT_RGB332:
            return 1;
            break;
        case IPU_PIX_FMT_RGB565:
        case IPU_PIX_FMT_YUYV:
        case IPU_PIX_FMT_UYVY:
            return 2;
            break;
        case IPU_PIX_FMT_BGR24:
        case IPU_PIX_FMT_RGB24:
            return 3;
            break;
        case IPU_PIX_FMT_GENERIC_32: //generic data
        case IPU_PIX_FMT_BGR32:
        case IPU_PIX_FMT_RGB32:
        case IPU_PIX_FMT_ABGR32:
            return 4;
            break;
        default:
            return 1;
            break;
    }
    return 0;
}

void _ipu_sdc_fg_init(ipu_channel_params_t * params)
{
    ulong reg;
    (void)params;

    // Enable FG & BG channel
	reg = IPU_REG(SDC_COM_CONF);
	IPU_REG(SDC_COM_CONF) = reg | SDC_COM_FG_EN | SDC_COM_BG_EN;
}

void _ipu_sdc_bg_init(ipu_channel_params_t * params)
{
    ulong reg;
    (void)params;

    // Enable BG channel
	reg = IPU_REG(SDC_COM_CONF);
	IPU_REG(SDC_COM_CONF) = reg | SDC_COM_BG_EN;
}

static void _ipu_ch_param_set_size(ulong * params, ulong pixel_fmt,
                                      ushort width, ushort height, 
                                      ushort stride)
{
    ulong u_offset = 0;
    ulong v_offset = 0;
    memset((void*)params, 0, 40);

    params[3] = (ulong)((width-1) << 12) | ((ulong)(height-1) << 24);
    params[4] = (ulong)(height-1) >> 8;
    params[7] = (ulong)(stride - 1) << 3;

    switch (pixel_fmt) {
        case IPU_PIX_FMT_GENERIC:
            //Represents 8-bit Generic data
            params[7] |= 3 | (7UL << (81-64)) | (31L << (89-64));   // BPP & PFS
            params[8] = 2;                      // SAT = use 32-bit access
            break;
        case IPU_PIX_FMT_GENERIC_32:
            //Represents 32-bit Generic data
            params[7] |= (7UL << (81-64)) | (7L << (89-64));   // BPP & PFS
            params[8] = 2;                      // SAT = use 32-bit access
            break;
        case IPU_PIX_FMT_RGB565:
            params[7] |= 2L | (4UL << (81-64)) | (7L << (89-64));        // BPP & PFS
            params[8] = 2 |                     // SAT = 32-bit access
                        (0UL  <<(99-96)) |      // Red bit offset
                        (5UL  <<(104-96)) |     // Green bit offset
                        (11UL <<(109-96)) |     // Blue bit offset
                        (16UL <<(114-96)) |     // Alpha bit offset
                        (4UL  <<(119-96)) |     // Red bit width - 1
                        (5UL  <<(122-96)) |     // Green bit width - 1
                        (4UL  <<(125-96));      // Blue bit width - 1
            break;
        case IPU_PIX_FMT_BGR24:
            params[7] |= 1 | (4UL << (81-64)) | (7L << (89-64)); // 24 BPP & RGB PFS
            params[8] = 2 |                     // SAT = 32-bit access
                        (8UL  <<(104-96)) |     // Green bit offset
                        (16UL <<(109-96)) |     // Blue bit offset
                        (24UL << (114-96)) |    // Alpha bit offset
                        (7UL  <<(119-96)) |     // Red bit width - 1
                        (7UL  <<(122-96)) |     // Green bit width - 1
                        (ulong)(7UL  <<(125-96));    // Blue bit width - 1
            break;
        case IPU_PIX_FMT_RGB24:
            params[7] |= 1 | (4UL << (81-64)) | (7L << (89-64)); // 24 BPP & RGB PFS
            params[8] = 2 |                     // SAT = 32-bit access
                        (16UL <<(99-96))  |     // Red bit offset
                        (8UL  <<(104-96)) |     // Green bit offset
                        (0UL  <<(109-96)) |     // Blue bit offset
                        (24UL << (114-96)) |    // Alpha bit offset
                        (7UL  <<(119-96)) |     // Red bit width - 1
                        (7UL  <<(122-96)) |     // Green bit width - 1
                        (ulong)(7UL  <<(125-96));    // Blue bit width - 1
            break;
        case IPU_PIX_FMT_BGRA32:
        case IPU_PIX_FMT_BGR32:
            params[7] |= 0 | (4UL << (81-64)) | (7 << (89-64)); // BPP & pixel fmt
            params[8] = 2 |                     // SAT = 32-bit access
                        (8UL  <<(99-96))  |     // Red bit offset
                        (16UL <<(104-96)) |     // Green bit offset
                        (24UL <<(109-96)) |     // Blue bit offset
                        (0UL  <<(114-96)) |     // Alpha bit offset
                        (7UL  <<(119-96)) |     // Red bit width - 1
                        (7UL  <<(122-96)) |     // Green bit width - 1
                        (ulong)(7UL  <<(125-96));    // Blue bit width - 1
            params[9] = 7;                      // Alpha bit width - 1
            break;
        case IPU_PIX_FMT_RGBA32:
        case IPU_PIX_FMT_RGB32:
            params[7] |= 0 | (4UL << (81-64)) | (7 << (89-64)); // BPP & pixel fmt
            params[8] = 2 |                     // SAT = 32-bit access
                        (24UL <<(99-96))  |     // Red bit offset
                        (16UL <<(104-96)) |     // Green bit offset
                        (8UL  <<(109-96)) |     // Blue bit offset
                        (0UL  <<(114-96)) |     // Alpha bit offset
                        (7UL  <<(119-96)) |     // Red bit width - 1
                        (7UL  <<(122-96)) |     // Green bit width - 1
                        (ulong)(7UL  <<(125-96));    // Blue bit width - 1
            params[9] = 7;                      // Alpha bit width - 1
            break;
        case IPU_PIX_FMT_ABGR32:
            params[7] |= 0 | (4UL << (81-64)) | (7 << (89-64)); // BPP & pixel fmt
            params[8] = 2 |                     // SAT = 32-bit access
                        (0UL  <<(99-96))  |     // Alpha bit offset
                        (8UL  <<(104-96)) |     // Blue bit offset
                        (16UL <<(109-96)) |     // Green bit offset
                        (24UL <<(114-96)) |     // Red bit offset
                        (7UL  <<(119-96)) |     // Alpha bit width - 1
                        (7UL  <<(122-96)) |     // Blue bit width - 1
                        (ulong)(7UL  <<(125-96));    // Green bit width - 1
            params[9] = 7;                      // Red bit width - 1
            break;
        case IPU_PIX_FMT_YUYV:
            params[7] |= 2 | (6UL << 17) | (7 << (89-64)); // BPP & pixel format
            params[8] = 2;                      // SAT = 32-bit access
            break;
        case IPU_PIX_FMT_YUV420P2:
        case IPU_PIX_FMT_YUV420P:
            params[7] |= 3 | (3UL << 17) | (7 << (89-64)); // BPP & pixel format
            params[8] = 2;                      // SAT = 32-bit access
            u_offset = stride * height;
            v_offset = u_offset + u_offset/4;
            break;
        case IPU_PIX_FMT_YVU422P:
            params[7] |= 3 | (2UL << 17) | (7 << (89-64)); // BPP & pixel format
            params[8] = 2;                      // SAT = 32-bit access
            v_offset = stride * height;
            u_offset = v_offset + v_offset/2;
            break;
        default:
			;
            break;
    }

    params[1] = (1UL << (46-32)) | (u_offset << (53-32));
    params[2] = u_offset >> (64-53);
    params[2] |= v_offset << (79-64);
    params[3] |= v_offset >> (96-79);
}

static void _ipu_ch_param_set_burst_size(ulong * params, ushort burst_pixels)
{
    params[7] &= ~(0x3FL << (89-64));
    params[7] |= (ulong)(burst_pixels - 1) << (89-64);
}    

static void _ipu_ch_param_set_buffer(ulong * params, void * buf0, void * buf1)
{
    params[5] = (ulong)buf0;
    params[6] = (ulong)buf1;
}

static void _ipu_ch_param_set_rotation(ulong * params, ipu_rotate_mode_t rot)
{
    params[7] |= (ulong)rot << (84-64);
}

void _ipu_write_param_mem(ulong addr, ulong * data, ulong numWords)
{
    for ( ; numWords > 0; numWords--)
    {
		IPU_REG(IPU_IMA_ADDR) = addr;
		IPU_REG(IPU_IMA_DATA) = *data++;
        addr++;
        if ((addr & 0x7) == 5)
        {
            addr &= ~0x7;       // set to word 0
            addr += 8;          // increment to next row
        }
    }
}

/*!
 * This function is called to initialize a buffer for logical IPU channel.
 * 
 * @param       channel         Input parameter for the logical channel ID.
 *
 * @param       type            Input parameter which buffer to initialize.
 *
 * @param       pixel_fmt       Input parameter for pixel format of buffer. Pixel 
 *                              format is a FOURCC ASCII code.
 *
 * @param       width           Input parameter for width of buffer in pixels.
 *
 * @param       height          Input parameter for height of buffer in pixels.
 *
 * @param       stride          Input parameter for stride length of buffer 
 *                              in pixels.
 *
 * @param       rot_mode        Input parameter for rotation setting of buffer.
 *                              A rotation setting other than \b IPU_ROTATE_VERT_FLIP 
 *                              should only be used for input buffers of rotation 
 *                              channels. 
 *
 * @param       phyaddr_0       Input parameter buffer 0 physical address.
 *
 * @param       phyaddr_1       Input parameter buffer 1 physical address.
 *                              Setting this to a value other than NULL enables
 *                              double buffering mode.
 *
 * @return      This function returns 0 on success or negative error code on fail
 */
ulong ipu_init_channel_buffer(ipu_channel_t channel, ipu_buffer_t type,
                                ulong pixel_fmt,
                                ushort width, ushort height,
                                ulong stride,
                                ipu_rotate_mode_t rot_mode,
                                void * phyaddr_0, void * phyaddr_1)
{
    ulong params[10];
    //ulong lock_flags;
    ulong reg;
    ulong dma_chan;
    ulong stride_bytes;

    dma_chan = channel_2_dma(channel, type);
    stride_bytes = stride * bytes_per_pixel(pixel_fmt);

    if (dma_chan == IDMA_CHAN_INVALID)
        return -1;

    if (stride_bytes % 4)
    {
        return -1;
    }

    // IC channels' stride must be multiple of 8 pixels
    if ((dma_chan <= 13) && (stride % 8))
    {
        return -1;
    }
    
    // Build parameter memory data for DMA channel
    _ipu_ch_param_set_size(params, pixel_fmt, width, height, stride_bytes);
    _ipu_ch_param_set_buffer(params, phyaddr_0, phyaddr_1);
    _ipu_ch_param_set_rotation(params, rot_mode);
    // Some channels (rotation) have restriction on burst length
    if ((dma_chan == 10) || (dma_chan == 11) || (dma_chan == 13))
    {
        _ipu_ch_param_set_burst_size(params, 8);
    }
    else if (dma_chan == 24)        // PF QP channel
    {
        _ipu_ch_param_set_burst_size(params, 4);
     }
    else if (dma_chan == 25)        // PF H264 BS channel
    {
        _ipu_ch_param_set_burst_size(params, 16);
    }
    else if (((dma_chan == 14) || (dma_chan == 15)) &&
             pixel_fmt == IPU_PIX_FMT_RGB565)
    {
        _ipu_ch_param_set_burst_size(params, 16);
    }

    _ipu_write_param_mem(DMAParamAddr(dma_chan), params, 10);

    reg = IPU_REG(IPU_CHA_DB_MODE_SEL);
    if (phyaddr_1 != NULL) {
        reg |= 1UL << dma_chan;
    }
    else {
        reg &= ~(1UL << dma_chan);
    }
	IPU_REG(IPU_CHA_DB_MODE_SEL) = reg;

    // Reset to buffer 0
	IPU_REG(IPU_CHA_CUR_BUF) = 1UL << dma_chan;

    return 0;
}

/*!
 * This function is called to set a channel's buffer as ready.
 * 
 * @param       channel         Input parameter for the logical channel ID.
 *
 * @param       type            Input parameter which buffer to initialize.
 *
 * @param       bufNum          Input parameter for which buffer number set to
 *                              ready state.
 *
 * @return      This function returns 0 on success or negative error code on fail
 */
ulong ipu_select_buffer(ipu_channel_t channel, ipu_buffer_t type, ulong bufNum)
{
    ulong dma_chan = channel_2_dma(channel, type);

    if (dma_chan == IDMA_CHAN_INVALID)
        return -1;

    if (bufNum == 0)
    {
        //Mark buffer 0 as ready.
		IPU_REG(IPU_CHA_BUF0_RDY) = 1UL << dma_chan;
    }
    else
    {
        //Mark buffer 1 as ready.
		IPU_REG(IPU_CHA_BUF1_RDY) = 1UL << dma_chan;
    }

    return 0;
}

/*!
 * This function enables a logical channel.
 * 
 * @param       channel         Input parameter for the logical channel ID.
 *
 * @return      This function returns 0 on success or negative error code on 
 *              fail. 
 */
ulong ipu_enable_channel( ipu_channel_t channel )
{
    ulong reg;
    //ulong lock_flags;
    ulong in_dma;
    ulong sec_dma;
    ulong out_dma;
    ulong chan_mask = 0;
	
	reg = IPU_REG(IDMAC_CHA_EN);

    // Get input and output dma channels
    out_dma = channel_2_dma(channel, IPU_OUTPUT_BUFFER);
    if (out_dma != IDMA_CHAN_INVALID)
        reg |= 1UL << out_dma;
    in_dma = channel_2_dma(channel, IPU_INPUT_BUFFER);
    if (in_dma != IDMA_CHAN_INVALID)
        reg |= 1UL << in_dma;

    // Get secondary input dma channel
    if (gSecChanEn[IPU_CHAN_ID(channel)])
    {
        sec_dma = channel_2_dma(channel, IPU_SEC_INPUT_BUFFER);
        if (sec_dma != IDMA_CHAN_INVALID)
        {
            reg |= 1UL << sec_dma;
        }
    }
	
	IPU_REG(IDMAC_CHA_EN) = reg | chan_mask;

    switch(channel)
    { 
        case MEM_SDC_BG:
             _ipu_sdc_bg_init(NULL);
            break;
        case MEM_SDC_FG:
            _ipu_sdc_fg_init(NULL);
            break;
        default:
            ;
            break;
    }

    return 0;
}

/*!
 * This function is called to initialize a synchronous LCD panel.
 * 
 * @param       panel           The type of panel.
 *
 * @param       refreshRate     Desired refresh rate for the panel in Hz. 
 *
 * @param       width           The width of panel in pixels.
 *
 * @param       height          The height of panel in pixels.
 *
 * @param       pixel_fmt       Input parameter for pixel format of buffer. Pixel 
 *                              format is a FOURCC ASCII code.
 *
 * @param       hStartWidth     The number of pixel clocks between the HSYNC
 *                              signal pulse and the start of valid data.
 *
 * @param       hSyncWidth      The width of the HSYNC signal in units of pixel 
 *                              clocks.
 *
 * @param       hEndWidth       The number of pixel clocks between the end of
 *                              valid data and the HSYNC signal for next line.
 *
 * @param       vStartWidth     The number of lines between the VSYNC
 *                              signal pulse and the start of valid data.
 *
 * @param       vSyncWidth      The width of the VSYNC signal in units of lines 
 *
 * @param       vEndWidth       The number of lines between the end of valid
 *                              data and the VSYNC signal for next frame.
 *
 * @param       sig             Bitfield of signal polarities for LCD interface.
 *
 * @return      This function returns 0 on success or negative error code on 
 *              fail.
 */
ulong ipu_sdc_init_panel(ipu_panel_t panel, 
                         uchar refreshRate,
                         ushort width, ushort height,
                         ulong pixel_fmt, 
                         ushort hStartWidth, ushort hSyncWidth, ushort hEndWidth, 
                         ushort vStartWidth, ushort vSyncWidth, ushort vEndWidth,
				                 ipu_di_signal_cfg_t sig)
{
    //ulong lock_flags;
    ulong reg;
    ulong old_conf;
    ulong div;
    ulong pixel_clk;
    
    if ((vSyncWidth == 0) || (hSyncWidth == 0))
        return -1;

    // Init panel size and blanking periods
    reg = ((ulong)(hSyncWidth - 1) << 26) | ((ulong)(width + hStartWidth + hEndWidth - 1) << 16);
  	IPU_REG(SDC_HOR_CONF) = reg;     // YHN: UM Rev 2.3 Section 44.3.3.6.8

    reg = ((ulong)(vSyncWidth - 1) << 26) | ((ulong)(height + vStartWidth + vEndWidth - 1) << 16 | SDC_V_SYNC_WIDTH_L);
	  IPU_REG(SDC_VER_CONF) = reg;    // YHN: UM Rev 2.3 Section 44.3.3.6.9

    g_hStartWidth = hStartWidth;
    g_vStartWidth = vStartWidth;

    switch (panel)
    {
        case IPU_PANEL_TFT:
//			IPU_REG(SDC_COM_CONF) = SDC_CONF_SHARP | SDC_COM_TFT_COLOR; // use this for OSD043TN14
			IPU_REG(SDC_COM_CONF) = SDC_COM_TFT_COLOR;	// use this for OSD043TN24
            break;
        default:
            return -1;
    }

	// Init clocking

    // Calculate total clocks per frame
    pixel_clk = (width + hStartWidth + hEndWidth) * (height + vStartWidth + vEndWidth) * refreshRate;

    // Calculate divider
    // fractional part is 4 bits so simply multiple by 2^4 to get fractional part
    //div = (g_ipu_clk * 16) / pixel_clk;
    
    // YHN assuming g_ipu_clk is 66500000
    g_ipu_clk = 133000000;
    div = (g_ipu_clk * 16) / pixel_clk; 
    if (div < 0x40) // Divider less than 4
    {
        div = 0x40;
    }

    // DISP3_IF_CLK_DOWN_WR is half the divider value and 2 less fraction bits
    // Subtract 1 extra from DISP3_IF_CLK_DOWN_WR based on timing debug
    // DISP3_IF_CLK_UP_WR is 0
    IPU_REG(DI_DISP3_TIME_CONF) = (((div/8) - 1) << 22) | div;
     //IPU_REG(DI_DISP3_TIME_CONF) = 0x02400084; 

    // DI settings
    old_conf = IPU_REG(DI_DISP_IF_CONF) & 0x78FFFFFF;
    old_conf |= sig.datamask_en << DI_D3_DATAMSK_SHIFT |
                sig.clksel_en << DI_D3_CLK_SEL_SHIFT |
                sig.clkidle_en << DI_D3_CLK_IDLE_SHIFT;
	  IPU_REG(DI_DISP_IF_CONF) = old_conf;

    old_conf = IPU_REG(DI_DISP_SIG_POL) & 0xE0FFFFFF;
    old_conf |= sig.enable_pol << DI_D3_DRDY_SHARP_POL_SHIFT; 
    //            sig.Hsync_pol << DI_D3_HSYNC_POL_SHIFT | 
    //            sig.Vsync_pol << DI_D3_VSYNC_POL_SHIFT;
	  IPU_REG(DI_DISP_SIG_POL) = old_conf;

    switch (pixel_fmt)
    {
        case IPU_PIX_FMT_RGB24:
			IPU_REG(DI_DISP3_B0_MAP) = DI_mappings[0];
			IPU_REG(DI_DISP3_B1_MAP) = DI_mappings[1];
			IPU_REG(DI_DISP3_B2_MAP) = DI_mappings[2];
			IPU_REG(DI_DISP_ACC_CC) = IPU_REG(DI_DISP_ACC_CC) | ((DI_mappings[3]-1)<<12);
            break;
        case IPU_PIX_FMT_RGB666:
			IPU_REG(DI_DISP3_B0_MAP) = DI_mappings[4];
			IPU_REG(DI_DISP3_B1_MAP) = DI_mappings[5];
			IPU_REG(DI_DISP3_B2_MAP) = DI_mappings[6];
			IPU_REG(DI_DISP_ACC_CC) = IPU_REG(DI_DISP_ACC_CC) | ((DI_mappings[7]-1)<<12);
            break;
        default:
			IPU_REG(DI_DISP3_B0_MAP) = DI_mappings[8];
			IPU_REG(DI_DISP3_B1_MAP) = DI_mappings[9];
			IPU_REG(DI_DISP3_B2_MAP) = DI_mappings[10];
			IPU_REG(DI_DISP_ACC_CC) = IPU_REG(DI_DISP_ACC_CC) | ((DI_mappings[11]-1)<<12);
            break;
    }

	return 0;
}

/*!
 * This function sets the window position of the foreground or background plane.
 * modes.
 * 
 * @param       channel         Input parameter for the logical channel ID.
 *
 * @param       x_pos           The X coordinate position to place window at.
 *                              The position is relative to the top left corner.
 *
 * @param       y_pos           The Y coordinate position to place window at.
 *                              The position is relative to the top left corner.
 *
 * @return      This function returns 0 on success or negative error code on fail
 */
ulong ipu_sdc_set_window_pos(ipu_channel_t channel, ushort x_pos, ushort y_pos)
{
    //x_pos += g_hStartWidth;
    //y_pos += g_vStartWidth;
   // YHN, need to fix code so for now hard-set the values
    //x_pos = 0x1A;
    //y_pos = 0x03;
 	IPU_REG(SDC_BG_POS) = SDC_BGXP(43) | SDC_BGYP(11);
 	IPU_REG(SDC_FG_POS) = SDC_FGXP(0) | SDC_FGYP(0);
  
//    if ( channel == MEM_SDC_BG )
//    {
//		IPU_REG(SDC_BG_POS) = (x_pos << 16) | y_pos;
//    }
//    else if ( channel == MEM_SDC_FG )
//    {
//		IPU_REG(SDC_FG_POS) = (x_pos << 16) | y_pos;
//    }
//    else
//    {
//        return -1;
//    }
//
    return 0;
}

//ulong ipu_sdc_set_brightness( uchar value )
//{
//	IPU_REG(SDC_PWM_CTRL) = 0x03000000UL | value << 16;
//
//    return 0;
//}

/*!
 * This function sets the foreground and background plane global alpha blending 
 * modes.
 * 
 * @param       enable          Boolean to enable or disable global alpha 
 *                              blending. If disabled, per pixel blending is used.
 *
 * @param       alpha           Global alpha value.
 *
 * @return      This function returns 0 on success or negative error code on fail
 */
ulong ipu_sdc_set_global_alpha(int enable, uchar alpha)
{
    ulong reg;

    if (enable)
    {
		//reg = IPU_REG(SDC_GW_CTRL) & 0x00FFFFFFL;
		reg = IPU_REG(SDC_GRAPH_WIND_CTRL) & 0x00FFFFFFL;
		//IPU_REG(SDC_GW_CTRL) = reg | ((ulong)alpha << 24);
		IPU_REG(SDC_GRAPH_WIND_CTRL) = reg | ((ulong)alpha << 24);
		
		reg = IPU_REG(SDC_COM_CONF);
		IPU_REG(SDC_COM_CONF) = reg | SDC_COM_GLB_A;
    }
    else
    {
		reg = IPU_REG(SDC_COM_CONF);
		IPU_REG(SDC_COM_CONF) = reg & ~SDC_COM_GLB_A;
    }    

    return 0;
}

/*!
 * This function sets the transparent color key for SDC graphic plane.
 * 
 * @param       channel         Input parameter for the logical channel ID.
 *
 * @param       enable          Boolean to enable or disable color key
 *
 * @param       colorKey        24-bit RGB color to use as transparent color key. 
 *
 * @return      This function returns 0 on success or negative error code on fail
 */
ulong ipu_sdc_set_color_key(ipu_channel_t channel, int enable, ulong colorKey )
{
    ulong reg, sdc_conf;
    //ulong lock_flags;

	sdc_conf = IPU_REG(SDC_COM_CONF);
    if (channel == MEM_SDC_BG) {
        sdc_conf &= ~SDC_COM_GWSEL;
    }
    else {
        sdc_conf |= SDC_COM_GWSEL;
    }

    if (enable)
    {
		//reg = IPU_REG(SDC_GW_CTRL) & 0xFF000000L;
		reg = IPU_REG(SDC_GRAPH_WIND_CTRL) & 0xFF000000L;
		//IPU_REG(SDC_GW_CTRL) = reg | (colorKey & 0x00FFFFFFL);
		IPU_REG(SDC_GRAPH_WIND_CTRL) = reg | (colorKey & 0x00FFFFFFL);
        sdc_conf |= SDC_COM_KEY_COLOR_G;
    }
    else
    {
        sdc_conf &= ~SDC_COM_KEY_COLOR_G;
    }
	IPU_REG(SDC_COM_CONF) = sdc_conf;
  	
    return 0;
}

