//==========================================================================
//
//      ads7846.c
//
// Author(s):   Michael Kelly - Cogent Computer Systems, Inc.
// Date:        03/06/03
// Description:	ADS7846 Interface routines for CSB535FS/CSB935FS
//				Modified from MC9328mxl version to use SPI1 
//
//==========================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "stddefs.h"
#include "mc9328mx21.h"
#include "cpu_gpio.h"
#include "ads7846.h"
#include "cli.h"
#include "umongpio.h"

//--------------------------------------------------------------------------
// function prototypes
//
int ads_init(void);
int ads_rd(uchar ads_ctl);

extern void udelay(int delay);

#define SPI_CTRL_BIT_CNT(_x_)	((_x_ & 0xf) << 0)	// bit count - 1
#define SPI_CTRL_POL_0			0					// see the MC9328MXL mamual
#define SPI_CTRL_POL_1			BIT4				// for a description of polarity
#define SPI_CTRL_PHASE_0		0					// and phase
#define SPI_CTRL_PHASE_1		BIT5				// 
#define SPI_CTRL_SS_ALWAYS		0					// SS stays low between transfers
#define SPI_CTRL_SS_PULSE		BIT6				// SS returns to inactive between transfers
#define SPI_CTRL_SS_LOW			0					// SS is low true
#define SPI_CTRL_SS_HIGH		BIT7				// SS is high true
#define SPI_CTRL_BUSY			BIT8				// read = transfer is in progress
#define SPI_CTRL_START			BIT8				// write = start transfer
#define SPI_CTRL_ENABLE			BIT9				// enable the SPI module
#define SPI_CTRL_MASTER			BIT10				// place us in Master mode
#define SPI_CTRL_RDY_OFF		(0x00 << 11)		// ignore RDY input
#define SPI_CTRL_RDY_EDGE		(0x01 << 11)		// wait for RDY input falling edge
#define SPI_CTRL_RDY_LEVEL		(0x02 << 11)		// wait for RDY input level low
#define SPI_CTRL_SCK4			(0x00 << 13)		// SCK divisor from PERCLK2 = 4	
#define SPI_CTRL_SCK8			(0x01 << 13)		// SCK divisor from PERCLK2 = 8	
#define SPI_CTRL_SCK16			(0x02 << 13)		// SCK divisor from PERCLK2 = 16	
#define SPI_CTRL_SCK32			(0x03 << 13)		// SCK divisor from PERCLK2 = 32	
#define SPI_CTRL_SCK64			(0x04 << 13)		// SCK divisor from PERCLK2 = 64	
#define SPI_CTRL_SCK128			(0x05 << 13)		// SCK divisor from PERCLK2 = 128
#define SPI_CTRL_SCK256			(0x06 << 13)		// SCK divisor from PERCLK2 = 256
#define SPI_CTRL_SCK512			(0x07 << 13)		// SCK divisor from PERCLK2 = 512

//--------------------------------------------------------------------------
// ads_init()
//
// This routine sets up the MC9328MX21 SPI port in Microwire mode
// (8 bit control, 16 bit data, 24 bit total).  It also turns on
// the ADS7846 pen interrupt via a dummy read.  Note that we assume
// that PERCLK2 is 12Mhz (HCLK/4).  
//
int ads_init()
{

	SPI1_REG(SPI_CTRL) = SPI_CTRL_BIT_CNT(7)		// 8-bit data mode
					   | SPI_CTRL_POL_1				// POL_1 and  PHASE_1 = SCLK parks high 
					   | SPI_CTRL_PHASE_1			// and rises in the middle of the data
					   | SPI_CTRL_SS_PULSE			// SS goes inactive between transfers
					   | SPI_CTRL_MASTER			// Master mode
					   | SPI_CTRL_RDY_OFF			// ignore SPI_RDY
					   | SPI_CTRL_ENABLE			// turn the module on
					   | SPI_CTRL_SCK128;			// 12Mhz(PERCLK2)/128 = approx. 94Khz

	// enable the ADS7846 so it can generate a touch interrupt.
	// this consists of reading any channel, but setting the
	// power down mode in the control byte to 00b.  note we
	// don't bother with the returned data
	if (ads_rd(ADS7846E_S | ADS7846E_PD_LPWR) == -1)
	{
		printf("Error returned from ads_rd(0x%02x)!\n", (ADS7846E_S | ADS7846E_PD_LPWR));
		return -1;
	}
	return 0;
}

//--------------------------------------------------------------------------
int ads_rd(uchar ads_ctl)
{
	int timeout = 100;
	uchar temp0, temp1;

	// the MC9328MXL only handles up to 16-bits per transfer
	// so we send 8-bits at a time to get our total of 24
	// enable the slave select
	GPIO_clr(TSEN);
	SPI1_REG(SPI_TXD) = ads_ctl;
	SPI1_REG(SPI_CTRL) |= SPI_CTRL_START;


	// wait for not busy, then read the data
	while (timeout)
	{
		udelay(500);	// 24bits at 128Khz takes about 250usec
		if ((SPI1_REG(SPI_CTRL) & SPI_CTRL_BUSY) == 0) break;
		timeout --;
	}
	
	if (timeout == 0) return -1;
	else 
	{
		// toss the first byte
		temp0 = SPI1_REG(SPI_RXD);

		SPI1_REG(SPI_TXD) = 0;
		SPI1_REG(SPI_CTRL) |= SPI_CTRL_START;
		while ((SPI1_REG(SPI_CTRL) & SPI_CTRL_BUSY));
		temp0 = SPI1_REG(SPI_RXD);

		SPI1_REG(SPI_TXD) = 0;
		SPI1_REG(SPI_CTRL) |= SPI_CTRL_START;
		while ((SPI1_REG(SPI_CTRL) & SPI_CTRL_BUSY));
		temp1 = SPI1_REG(SPI_RXD);

		// disable the slave select
		GPIO_set(TSEN);

//		return (((temp0 << 8) | temp1) >> 4);	// lower 4-bit are padding
		return ((temp0 << 8) | temp1);
	}

	return 0;
}

//--------------------------------------------------------------------------
char *adsHelp[] = {
	"Screen touch demo using ADS7846 and MC9328MXL SPI port.",
	"",
	0
};

int ads(int argc, char *argv[])
{
	uchar c;
	int pen, i;
	int average_x, average_y;

	// init the SPI and ADS7846
	if (ads_init() == -1)
	{
		printf("Error intializing ADS7846!\n");
		return (CMD_FAILURE);
	}

	printf("Waiting for Touch (Press \'X\' to end test)...\n");

	pen = 0;
	while (1)
	{
		if (gotachar())
		{
			c = getchar();
			if ((c == 'X') || (c == 'x')) goto done;
		}
		if (GPIO_tst(PIRQ) == 0)
		{
			printf("Pen Down....\n");
			pen = 1;
			while(GPIO_tst(PIRQ) == 0) // keep reading until touch goes away
			{
				average_x = 0;
				average_y = 0;

				// display every 256 samples (less if pen is down a short time)
				for (i = 0; i < 256 ; i++)
				{
					average_x += ads_rd(ADS7846E_S | ADS7846E_PD_LPWR | ADS7846E_ADD_DFR_X);
					average_y += ads_rd(ADS7846E_S | ADS7846E_PD_LPWR | ADS7846E_ADD_DFR_Y);
				}
				printf("X = %04d, Y = %04d\n", average_x/i, average_y/i);

			} // while pen is down
		}
		if (pen)
		{
			printf("Pen Up....\n");
			pen = 0;
		}
	}

done:		
	return(CMD_SUCCESS);
}
