//=============================================================================
//
//      cpuio.c
//
//      CPU/Board Specific IO
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         09/18/2005
// Description:  This file contains the IO functions required by Micro Monitor
//				 that are unique to the CSB535FS
//
//
//=============================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "stddefs.h"
#include "warmstart.h"
#include "mc9328mx21_uart.h"
#include "mc9328mx21.h"
#include "mc9328mx21_lcd.h"
#include "mc9328mx21_mem.h"
#include "cpu_gpio.h"
#include "fb_draw.h"
#include "umongpio.h"

extern ulong i2c_init(void);
extern ulong getpsr(void);
extern void putpsr(ulong);

// globals to hold background and foreground colors
uchar bcolor=0;		// vga black
//uchar color=15;		// vga white

// Delay for some usecs. - Not accurate, assumes ROM mode
// and no Cache
void udelay(int delay)
{
  volatile int i;
  for ( i = LOOPS_PER_USEC * delay; i ; i--)
	;
}


/* InitUART():
 * This is the top-level UART initialization for the monitor. 
 */
void
InitUART(int baud)
{
    if (baud == 0)
        baud = ConsoleBaudRate;

    /* Keep track of the last baud rate, so that it can be used if the */
    /* incoming baudrate is NULL. */
    ConsoleBaudRate = baud;

	// reset tx and rx
	UART4_REG(UART_CR2) = 0;

	// clear any pending interrupts and mask all further interrupts
	UART4_REG(UART_SR1) = 0xffffffff;
	UART4_REG(UART_SR2) = 0xffffffff;
	UART4_REG(UART_CR1) = 0x00000000;
	UART4_REG(UART_CR4) = 0x00000000;

	// set the RX mux to Uart, not IrDA - this is different than i.MXL
	UART4_REG(UART_CR3) = UART_CR3_RXDMUXSEL;

	// enable the uart and it's clocks - no auto baud, no interrupts
	UART4_REG(UART_CR1) |= UART_CR1_UARTEN;

	//enable the uart clocks
	PLL_REG(PLL_PCCR0) = 0x0000000f;

	// set the word parameters
	UART4_REG(UART_CR2) = UART_CR2_WS 	// 8-bit, 1 stop, no parity
					| UART_CR2_IRTS		// ignore RTS 
					| UART_CR2_SRST;	// no reset

	// set the fifo's to legal interrupt threshold values - unused by us
	// set input clock divisor to 1
	UART4_REG(UART_FCR) = UART_FCR_TXTL(2)
					| UART_FCR_RFDIV1
					| UART_FCR_RXTL(1);

	// set the baud rate denominator to 10,000.  Then divide
	// the desired baud by 100 and put that into the numerator.
	// This works when using a reference clock of 16Mhz with
	// RFDIV set to 1.  It also only works down to 1200 baud,
	// but who cares!
	UART4_REG(UART_BIR) = (baud/100) - 1;	// numerator
	UART4_REG(UART_BMR) = 10000 - 1;		// denominator

	// Enable TX and RX, release SRST
	UART4_REG(UART_CR2) |= UART_CR2_TXEN
					| UART_CR2_RXEN
					| UART_CR2_SRST;

    InitRemoteIO();
}

int
ConsoleBaudSet(int baud)
{
	// set the baud rate denominator to 10,000.  Then divide
	// the desired baud by 100 and put that into the numerator.
	// This works when using a reference clock of 16Mhz with
	// RFDIV set to 1.  It also only works down to 1200 baud,
	// but who cares!
	UART4_REG(UART_BIR) = (baud/100) - 1;	// numerator
	UART4_REG(UART_BMR) = 10000 - 1;		// denominator
	return(0);
}

int
devInit(int baud)
{
	// disable all perpihperal clocks.  They ahve to enable
	// themselves
	PLL_REG(PLL_PCCR0) = 0;

	// Write FMCR register in order to allow UART4 RXD to come out on PB29.
	SYS_REG(SYS_FMCR) = 0xfdffffcb;

	// Disable all GPIO for now.  We will allow the gpio_init()
	// code to do it later.
	GPIOA_REG(GPIO_GIUS) = 0;	// no GPIO on Port A for now
	GPIOA_REG(GPIO_GPR) = 0;
	GPIOB_REG(GPIO_GIUS) = 0;	// no GPIO on Port B for now
	GPIOB_REG(GPIO_GPR) = 0;
	GPIOC_REG(GPIO_GIUS) = 0;	// no GPIO on Port C for now
	GPIOC_REG(GPIO_GPR) = 0;
	GPIOD_REG(GPIO_GIUS) = 0;	// no GPIO on Port D for now
	GPIOD_REG(GPIO_GPR) = 0;
	GPIOE_REG(GPIO_GIUS) = 0;	// no GPIO on Port E for now
	GPIOE_REG(GPIO_GPR) = 0;
	GPIOF_REG(GPIO_GIUS) = 0;	// no GPIO on Port F for now
	GPIOF_REG(GPIO_GPR) = 0;

	// Initialize GPIO
    GPIO_init();	

    // Initialize the UART
    InitUART(baud);


//
// change the memory and i/o device timing to work with
// the modified 88.66mhz bus clock - cs0 and cs1
// cs5 for the cf card not tested
//

#if 1
 	// Setup CS0 for 110ns Spansion Flash
    EIM_REG(EIM_CS0U) = EIM_CSU_CNC_1		// hold CS negated one clock at the end of all accesses
					  | EIM_CSU_WSC(10)		// wait = 132ns (10+1 x 12ns, assumes BCLK = 88Mhz)
//					  | EIM_CSU_WSC(16)		// wait = 200ns (16+1 x 12ns, assumes BCLK = 88Mhz)
					  | EIM_CSU_WWS(2)		// additional wait states for writes
					  | EIM_CSU_EDC(2);		// 1 clock bus turn around time

	EIM_REG(EIM_CS0L) = EIM_CSL_OEA(0)		// CS asserted # 1/2 clocks before OE
					  | EIM_CSL_OEN(0)		// CS held # 1/2 clocks after OE
					  | EIM_CSL_WEA(0)		// CS asserted # 1/2 clocks before WE
					  | EIM_CSL_WEN(0)		// CS held # 1/2 clocks after WE
//					  | EIM_CSL_CSA(0)		// setup and hold address 2 clocks around CS
					  | EIM_CSL_CSA(1)		// setup and hold address 2 clocks around CS
					  | EIM_CSL_EBC			// assert EB0-3 for writes only
					  | EIM_CSL_DSZ_HALF_0	// 16-bits on D0-15
					  | EIM_CSL_CSEN;		// enable it

#endif
 
	// Setup CS1 for LAN9211
    EIM_REG(EIM_CS1U) = EIM_CSU_WSC(28);

	EIM_REG(EIM_CS1L) = EIM_CSL_OEA(1)		// CS asserted # 1/2 clocks before OE
			  		  | EIM_CSL_OEN(1)		// CS held # 1/2 clocks after OE
			  		  | EIM_CSL_WEA(1)		// CS asserted # 1/2 clocks before WE
			  		  | EIM_CSL_WEN(1)		// CS held # 1/2 clocks after WE
			  		  | EIM_CSL_CSA(2)		// setup and hold address 2 clocks around CS
			  		  | EIM_CSL_EBC			// assert EB0-3 for writes only
			  		  | EIM_CSL_DSZ_HALF_0  // 16-bits on D0-15
			  		  | EIM_CSL_CSEN;		// enable it

	// Initialize the I2C Port
	i2c_init();

	// Initialize the LCD controller
// 	lcd_color_depth = 16;		// 4, 8 or 16
//	lcd_init();
//	lcd_on();

	// Turn on USR_LED (GPIO 0)
//	GPIO_set(GPIO0);
//	GPIO_clr(GPIO0);
//	GPIO_set(GPIO0);

//	GPIO_clr(GPIO0);
//	GPIO_set(GPIO0);
//	GPIO_clr(GPIO0);

	return 0;
}

int
target_console_empty(void)
{
	if (UART4_REG(UART_SR2) & UART_SR2_TXFE)
		return(1);
	return(0);
}

int
target_putchar(char c)
{
    /* Wait for transmit ready bit, then send character.
	 */
    while(!(UART4_REG(UART_SR2) & UART_SR2_TXFE)){}
	UART4_REG(UART_TXD) = (ulong)c;
    return((int)c);
}

int
target_gotachar(void)
{

    if (UART4_REG(UART_SR2) & UART_SR2_RDR)
        return(1);
    return(0);
}

int 
target_getchar(void)
{
    char    c;


    c = (uchar)UART4_REG(UART_RXD) & 0xff;
    return((int)c);
}

void 
target_reset(void)
{
	warmstart(INITIALIZE);
}

void
initCPUio()
{
}

void
intsrestore(psr)
ulong     psr;
{
    putpsr(psr);
}

/*
 * Read the program status register (CPSR)
 * and set the FIQ and IRQ bits.
 */
ulong
intsoff(void)
{
    ulong  psr;

    psr = getpsr();

    /*
     * Set bit 6, bit 7 to disable interrupts.
     */
    putpsr(psr | 0x000000c0);
    return(psr);
}

/* target_timer():
 * Used in conjunction with INCLUDE_HWTMR and TIMER_TICKS_PER_MSEC
 * to set up a hardware based time base.
 */
unsigned long
target_timer(void)
{
	volatile unsigned register ccr;

   	asm("   MRC p15, 0, %0, c9, c13, 0" : "=r" (ccr));

	return(ccr);
}

/* cacheInitForTarget():
    Enable instruction cache only...
*/
void
cacheInitForTarget()
{
    asm("   MRC p15, 0, r0, c1, c0, 0");
    asm("   ORR r0, r0, #0x1000");  /* bit 12 is ICACHE enable*/
    asm("   MCR p15, 0, r0, c1, c0, 0");

    /* Flush instruction cache */
    asm("   MCR p15, 0, r0, c7, c5, 0");
}
