/*
 * etherdev.c:
 * Ethernet hookup for i.MX35's FEC to uMon.  Slightly modified (by Cogent)
 * from the mcf5329 port for use on CSB760.
 *
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 *
 * Author:	Ed Sutter
 * email:	esutter@alcatel-lucent.com
 * phone:	908-582-2351
 */

#include "config.h"
#include "genlib.h"
#include "stddefs.h"
#include "ether.h"
#include "mpc5121e.h"
#include "nbuf.h"
#include "cpuio.h"
#include "cache.h"
#include "timer.h"
#include "phy.h"
#include "endian.h"

#if INCLUDE_ETHERNET

#define MAX_ETH_PKT 1518
#define PHYADDR	31

#define FEC_EIR_ALL_EVENTS   \
   ( FEC_EIR_HBERR | FEC_EIR_BABR | FEC_EIR_BABT | \
     FEC_EIR_GRA   | FEC_EIR_TXF  | FEC_EIR_TXB  | \
     FEC_EIR_RXF   | FEC_EIR_RXB  | FEC_EIR_MII  | \
     FEC_EIR_EBERR | FEC_EIR_LC   | FEC_EIR_RL   | \
     FEC_EIR_UN )

#define FEC_RxBD_ERRS   \
   ( FEC_RxBD_LG | FEC_RxBD_NO |  FEC_RxBD_CR | \
     FEC_RxBD_OV | FEC_RxBD_TR ) 

#if ((RX_BUFFER_SIZE < MAX_ETH_PKT) || (TX_BUFFER_SIZE < MAX_ETH_PKT))
#error "MAX_ETH_PKT is too large."
#endif

int
phy_write(int phyadd, int regNumber, unsigned short value)
{
    struct	elapsed_tmr tmr;

	FEC_MMFR = 
		FEC_MMFR_ST(1) |
		FEC_MMFR_OP(1) |		// Write
		FEC_MMFR_PA(phyadd) |
		FEC_MMFR_RA(regNumber) |
		FEC_MMFR_TA(2) |
		FEC_MMFR_DATA(value);

    startElapsedTimer(&tmr,2500);
	while((FEC_EIR & FEC_EIR_MII) == 0) {
    	if(msecElapsed(&tmr))
			return(-1);
	}

	FEC_EIR = FEC_EIR_MII;
	return(0);
}

int
phy_read(int phyadd, int regNumber, unsigned short *value)
{
    struct	elapsed_tmr tmr;

	FEC_MMFR = 
		FEC_MMFR_ST(1) |
		FEC_MMFR_OP(2) |		// Read
		FEC_MMFR_PA(phyadd) |
		FEC_MMFR_RA(regNumber) |
		FEC_MMFR_TA(2);

    startElapsedTimer(&tmr,2500);
	while((FEC_EIR & FEC_EIR_MII) == 0) {
    	if(msecElapsed(&tmr))
			return(-1);
	}
	FEC_EIR = FEC_EIR_MII;
	*value = (unsigned short)(FEC_MMFR & 0xffff);
	return(0);
}

/* phy_reset():
 * Reset the phy.
 */
int
phy_reset(int phyadd)
{
	ushort ctrl;
	struct elapsed_tmr tmr;

	phy_read(phyadd,PHY_CTRL,&ctrl);
	ctrl |= PHY_CTRL_RST;
	phy_write(phyadd,PHY_CTRL,ctrl);
	startElapsedTimer(&tmr,10);
	while(1) {
		if (phy_read(phyadd,PHY_CTRL,&ctrl) < 0)
			return(-1);
		if ((ctrl & PHY_CTRL_RST) == 0)
			break;
		if (msecElapsed(&tmr)) {
			printf("phy reset timeout\n");
			return(-1);
		}
	}
	return(0);
}


/* phy_autoneg():
 * Go through the auto-negotiation process...
 */
int
phy_autoneg(int phy)
{
	volatile int retries;
	struct elapsed_tmr tmr;
	ushort ctrl, stat, autoneg, par;

#if INCLUDE_ETHERVERBOSE
	if (EtherVerbose &  SHOW_DRIVER_DEBUG)
		printf("\nPHY Autonegotiate... ");
#endif

	retries = 1000;
	autoneg = PHY_ANAD_802_3;

	phy_read(phy,PHY_STAT,&stat);
	if((stat & PHY_STAT_ANABLE) == 0) {
		printf("phy does not support autoneg\n");
		return (-1);
	}
	
	/* Set appropriate advertisement bits based on this
	 * phy's capabilities...
	 */
	if (stat & PHY_STAT_100FDX)
		autoneg |= PHY_ANAD_100FDX;
	else if (stat & PHY_STAT_10FDX)
		autoneg |= PHY_ANAD_10FDX;
	else if (stat & PHY_STAT_100HDX)
		autoneg |= PHY_ANAD_100HDX;
	else if (stat & PHY_STAT_10HDX)
		autoneg |= PHY_ANAD_10HDX;
	phy_write(phy,PHY_ANAD, autoneg);

	phy_read(phy,PHY_CTRL,&ctrl);
	ctrl &= ~(PHY_CTRL_LBK | PHY_CTRL_PWRDN);
	ctrl &= ~(PHY_CTRL_ISOLATE | PHY_CTRL_COLTST);
	ctrl |= (PHY_CTRL_ANE);

	phy_write(phy,PHY_CTRL, ctrl);
	ctrl |= PHY_CTRL_ANRESTART; 
	phy_write(phy,PHY_CTRL, ctrl);

	/* allow 10 seconds for auto-negotiation to complete...
	 */
	startElapsedTimer(&tmr,10000);
	do {
		phy_read(phy,PHY_STAT,&stat);
		if((stat & PHY_STAT_REMFAULT) != 0) {
			printf("phy: remote fault\n");
			return -1;
		}
		if (msecElapsed(&tmr)) {
#if INCLUDE_ETHERVERBOSE
			if (EtherVerbose &  SHOW_DRIVER_DEBUG)
				printf("phy: AN timeout\n");
#endif
			return(-1);
		}
	} while((stat & PHY_STAT_ANDONE) == 0);

	phy_read(phy,PHY_ANLPAR,&par);

	if((par & PHY_ANLP_REMFAULT) != 0) {
		printf("phy: remote fault\n");
		return -1;
	}

	phy_read(phy,PHY_CTRL,&ctrl);
	ctrl &= (~PHY_CTRL_FDX | PHY_CTRL_100MB);

	phy_read(phy,PHY_STAT,&stat);
	if ((stat & PHY_STAT_100FDX) && (par & PHY_ANLP_100FDX))
		ctrl |= (PHY_CTRL_100MB | PHY_CTRL_FDX);
	else if ((stat & PHY_STAT_10FDX) && (par & PHY_ANLP_10FDX))
		ctrl |= PHY_CTRL_FDX;
	else if ((stat & PHY_STAT_100HDX) && (par & PHY_ANLP_100HDX))
		ctrl |= PHY_CTRL_100MB;

	phy_write(phy,PHY_CTRL, ctrl);

#if INCLUDE_ETHERVERBOSE
	if (EtherVerbose &  SHOW_DRIVER_DEBUG) {
		printf("%dMbps %s-duplex\n",
			ctrl & PHY_CTRL_100MB ? 100 : 10,
			ctrl & PHY_CTRL_FDX ? "Full" : "Half");
	}
#endif
	return 0;
}

/*
 * enreset():
 *	Reset the PHY and MAC.
 */
void
enreset(void)
{
    struct	elapsed_tmr tmr;

	/* Reset the FEC - equivalent to a hard reset */
	FEC_ECR = FEC_ECR_RESET;

	/* Wait for the reset sequence to complete */
    startElapsedTimer(&tmr,2500);
	while(FEC_ECR & FEC_ECR_RESET) {
    	if(msecElapsed(&tmr)) {
			printf("enreset failed\n");
			return;
		}
	}

	/* Disable all FEC interrupts by clearing the IMR register */
	FEC_EIMR = 0;

	/* Clear any interrupts by setting all bits in the EIR register */
	FEC_EIR = FEC_EIR_ALL_EVENTS;
}

/*
 * eninit():
 * Initialize the PHY and MAC.
 * This would include establishing buffer descriptor tables and
 * all the support code that will be used by the ethernet device.
 *
 * It can be assumed at this point that the array uchar BinEnetAddr[6]
 * contains the 6-byte MAC address.
 *
 * Return 0 if successful; else -1.
 */
int
eninit(void)
{
	vuint32 *mibdat;

	/* set endian mode */
	FEC_DMACTRL = 0xC0000000;

	/* Disable all FEC interrupts by clearing the IMR register */
	FEC_EIMR = 0;

	/* Clear any interrupts by setting all bits in the EIR register */
	FEC_EIR = 0xffffffff;

	/* Set the source address for the controller */
	FEC_PALR =
		FEC_PALR_PADDR1((BinEnetAddr[0] << 24) | (BinEnetAddr[1] <<16)	
		| (BinEnetAddr[2] <<8) | (BinEnetAddr[3] <<0));

	FEC_PAUR = FEC_PAUR_PADDR2(
		(BinEnetAddr[4] <<8) | (BinEnetAddr[5] <<0)) |
		FEC_PAUR_TYPE(0x8808);

	/* Initialize the hash table registers */
	FEC_IAUR = 0;
	FEC_IALR = 0;

	FEC_GAUR = 0;
	FEC_GALR = 0;

	/* Set receive buffer size */
	FEC_EMRBR = MAX_ETH_PKT;

#if defined(USE_FULL_DUPLEX)
	FEC_RCR = FEC_RCR_MAX_FL(MAX_ETH_PKT) |
					FEC_RCR_MII_MODE;
	FEC_TCR = FEC_TCR_FDEN;
#else
	FEC_RCR = FEC_RCR_MAX_FL(MAX_ETH_PKT) |
					FEC_RCR_MII_MODE      |
					FEC_RCR_DRT;
	FEC_TCR = 0;
#endif

	/* Set MII speed to <= 2.5 MHz */
//	FEC_MSCR = FEC_MSCR_MII_SPEED(((BUS_FREQUENCY + 5000000 - 1) / 5000000));
	FEC_MSCR = FEC_MSCR_MII_SPEED(0x0E);

	/* Initialize the network buffers */
	nbuf_init();

	/* Point to the start of the circular Rx buffer descriptor queue */
	FEC_ERDSR = nbuf_get_start(Rx);

	/* Point to the start of the circular Tx buffer descriptor queue */
	FEC_ETDSR = nbuf_get_start(Tx);


#if defined(USE_FULL_DUPLEX)
	/* Set PHYS to 100 Mb/s, full duplex */
	phy_write(PHYADDR, 0, 0x3100);
#else
	/* Set PHYS to 100 Mb/s, half duplex */
	phy_write(PHYADDR, 0, 0x3000);
#endif

	FEC_FRSR = FEC_FRSR_R_FSTART(0x40);

	FEC_MIBC = FEC_MIBC_MIB_DISABLE;
	mibdat = &FEC_RMON_T_DROP;
	while (mibdat <= &FEC_IEEE_R_OCTETS_OK)
		*mibdat++ = 0;
	FEC_MIBC = 0;	

	/* Enable FEC */
	FEC_ECR = FEC_ECR_ETHER_EN;
	
	/* Indicate that there have been empty receive buffers produced */
	FEC_RDAR = FEC_RDAR_R_DES_ACTIVE;

	return (0);
}

int
EtherdevStartup(int verbose)
{
	/* Put ethernet controller in reset: */
	enreset();

	/* Initialize controller and return the value returned by
	 * eninit().
	 */
	return(eninit());
}

/* disablePromiscuousReception():
 * Provide the code that disables promiscuous reception.
 */
void
disablePromiscuousReception(void)
{
	FEC_RCR |= FEC_RCR_PROM;
}

/* enablePromiscuousReception():
 * Provide the code that enables promiscuous reception.
 */
void
enablePromiscuousReception(void)
{
	FEC_RCR &= ~FEC_RCR_PROM;
}

/* disableBroadcastReception():
 * Provide the code that disables broadcast reception.
 */
void
disableBroadcastReception(void)
{
	FEC_RCR |= FEC_RCR_BC_REJ;
}

/* enableBroadcastReception():
 * Provide the code that enables broadcast reception.
 */
void
enableBroadcastReception(void)
{
	FEC_RCR &= ~FEC_RCR_BC_REJ;
}

static	uchar testmsg[] = {
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0x08, 0x00, 't', 'e', 's', 't'
};

/* 
 * enselftest():
 *	Run a self test of the ethernet device(s).  This can be stubbed
 *	with a return(1).
 *	Return 1 if success; else -1 if failure.
 */
int
enselftest(int verbose)
{
	uchar	*msg;
	NBUF	*pNbuf;
	int		retval;
	ushort	id1, id2, phyreg;
    struct	elapsed_tmr tmr;

	retval = 1;

	/* According to section 23.5.10 of the i.MX35 reference manual...
	 * For both internal and external loopbacks, set FDEN to 1.
	 * For external,  clear both DRT & LOOP.
	 * For internal,  clear DRT & set LOOP.
	 */
	FEC_TCR |= FEC_TCR_FDEN;
	FEC_RCR &= ~(FEC_RCR_DRT | FEC_RCR_LOOP);

	if (getenv("EN_INT_LBK")) {
		if (verbose)
			printf("Self test (MAC-based lbk)...");

		FEC_RCR |= FEC_RCR_LOOP;
	}
	else {
		if (verbose)
			printf("Self test (PHY-based lbk)...");

		/* First read the PHY ID registers and verify valid device ID.
		 */
		if ((phy_read(PHYADDR,2,&id1) == -1) ||
			(phy_read(PHYADDR,3,&id2) == -1)) {
			printf("Phy access error\n");
			return(-1);
		}
		if ((id1 != 0x0007) || (id2 != 0xc0c4)) 
			printf("WARNING: unexpected PHY-ID %04x %04x\n",id1,id2);

		/* Next, put the PHY in loopback and send a packet to see if we receive
		 * what we transmit...
		 */
		phy_read(PHYADDR,PHY_CTRL,&phyreg);
		if (phy_write(PHYADDR,PHY_CTRL,phyreg | PHY_CTRL_LBK) == -1) {
			if (verbose)
				printf("phy write failed\n");
			return(-1);
		}
		phy_read(PHYADDR,PHY_CTRL,&phyreg);
	}

	msg = getXmitBuffer();
	memcpy((char *)msg,(char *)testmsg,sizeof(testmsg));
	sendBuffer(64);

	/* Wait for the looped back message... */

    startElapsedTimer(&tmr,2500);
	do {
    	if(msecElapsed(&tmr)) {
			printf("lbk pkt not rcvd\n");
			retval = -1;
			goto done;
		}

	} while(!nbuf_rx_next_ready());

	/* Get the next buffer received by the FEC */
	pNbuf = nbuf_rx_allocate();
	if (pNbuf == 0) {
		printf("rx_allocate1 failed\n");
		retval = -1;
		goto done;
	}
	if (!(pNbuf->status & FEC_RxBD_L))
	{
		do
		{	/* Free the current buffer */
			nbuf_rx_release(pNbuf);

			/* Get the next buffer of this frame */
			pNbuf = nbuf_rx_allocate();
			if (pNbuf == 0) {
				printf("rx_allocate2 failed\n");
				retval = -1;
				goto done;
			}
	
			/* Keep going until we find the last one */
		} while (!(pNbuf->status & FEC_RxBD_L));

		/* Free the last buffer of this frame */
		nbuf_rx_release(pNbuf);
		printf("rx_allocate3 failed\n");
		retval = -1;
		goto done;
	}
	
	nbuf_rx_release(pNbuf);
	FEC_RDAR = FEC_RDAR_R_DES_ACTIVE;

	if (memcmp((char *)pNbuf->data,(char *)testmsg,sizeof(testmsg))) {
		printf("lbk failed\n");
		retval = -1;
		goto done;
	}

	if (verbose)
		printf("passed\n");

	/* Put things back to normal... */
done:
	enreset();
	eninit();
	return(retval);
}

/* ShowEtherdevStats():
 * This function is used to display device-specific stats (error counts
 * usually).
 */
void
ShowEtherdevStats(void)
{
	ushort phyreg;

	/* Read the register twice because the STAT bit is sticky,
	 * and we want to make sure that we are showing the current status.
	 */
	phy_read(PHYADDR,PHY_STAT,&phyreg);
	phy_read(PHYADDR,PHY_STAT,&phyreg);
	if (phyreg & PHY_STAT_LINKUP) {
		phy_read(PHYADDR,PHY_CTRL,&phyreg);
		printf("PHY: %dMbps, %s-duplex\n",
			phyreg & PHY_CTRL_100MB ? 100 : 10,
			phyreg & PHY_CTRL_FDX ? "Full" : "Half");
	}
	else {
		printf("PHY: link is down\n");
		return;
	}
	printf(" Rx Packet Count:%-8lu", FEC_RMON_R_PACKETS);
	printf("    Rx Broadcast:%-8lu\n", FEC_RMON_R_BC_PKT);
	printf("    Rx Multicast:%-8lu", FEC_RMON_R_MC_PKT);
	printf(" CRC/Align error:%-8lu", FEC_RMON_R_CRC_ALIGN);
	printf("    Rx Undersize:%-8lu\n", FEC_RMON_R_UNDERSIZE);
	printf("     Rx Oversize:%-8lu", FEC_RMON_R_OVERSIZE);
	printf("     Rx Fragment:%-8lu", FEC_RMON_R_FRAG);
	printf("       Rx Jabber:%-8lu\n", FEC_RMON_R_JAB);
	printf("           Rx 64:%-8lu", FEC_RMON_R_P64);
	printf("       Rx 65-127:%-8lu", FEC_RMON_R_P65TO127);
	printf("      Rx 128-255:%-8lu\n", FEC_RMON_R_P128TO255);
	printf("      Rx 256-511:%-8lu", FEC_RMON_R_P256TO511);
	printf("     Rx 511-1023:%-8lu", FEC_RMON_R_512TO1023);
	printf("    Rx 1024-2047:%-8lu\n", FEC_RMON_R_1024TO2047);
	printf("       Rx >=2048:%-8lu", FEC_RMON_R_P_GTE2048);
	printf("       Rx Octets:%-8lu", FEC_RMON_R_OCTETS);
	printf("      Rx Dropped:%-8lu\n", FEC_IEEE_R_DROP);
	printf("     Rx frame OK:%-8lu", FEC_IEEE_R_FRAME_OK);
	printf("    Rx CRC error:%-8lu", FEC_IEEE_R_CRC);
	printf("  Rx Align error:%-8lu\n", FEC_IEEE_R_ALIGN);
	printf("   FIFO Overflow:%-8lu", FEC_IEEE_R_MACERR);
	printf(" Rx Pause Frames:%-8lu", FEC_IEEE_R_FDXFC);
	printf("    Rx Octets OK:%-8lu\n", FEC_IEEE_R_OCTETS_OK);
	printf("  Tx Unaccounted:%-8lu", FEC_RMON_T_DROP);
	printf(" Tx Packet Count:%-8lu", FEC_RMON_T_PACKETS);
	printf("    Tx Broadcast:%-8lu\n", FEC_RMON_T_BC_PKT);
	printf("    Tx Multicast:%-8lu", FEC_RMON_T_MC_PKT);
	printf(" CRC/Align error:%-8lu", FEC_RMON_T_CRC_ALIGN);
	printf("    Tx Undersize:%-8lu\n", FEC_RMON_T_UNDERSIZE);
	printf("     Tx Oversize:%-8lu", FEC_RMON_T_OVERSIZE);
	printf("     Tx Fragment:%-8lu", FEC_RMON_T_FRAG);
	printf("       Tx Jabber:%-8lu\n", FEC_RMON_T_JAB);
	printf("   Tx Collisions:%-8lu", FEC_RMON_T_COL);
	printf("           Tx 64:%-8lu", FEC_RMON_T_P64);
	printf("       Tx 65-127:%-8lu\n", FEC_RMON_T_P65TO127);
	printf("      Tx 128-255:%-8lu", FEC_RMON_T_P128TO255);
	printf("      Tx 256-511:%-8lu", FEC_RMON_T_P256TO511);
	printf("     Tx 511-1023:%-8lu\n", FEC_RMON_T_P512TO1023);
	printf("    Tx 1024-2047:%-8lu", FEC_RMON_T_P1024TO2047);
	printf("       Tx >=2048:%-8lu", FEC_RMON_T_P_GTE2048);
	printf("       Tx Octets:%-8lu\n", FEC_RMON_T_OCTETS);
	printf("      Tx Dropped:%-8lu", FEC_IEEE_T_DROP);
	printf("     Tx Frame OK:%-8lu", FEC_IEEE_T_FRAME_OK);
	printf("  Tx 1 Collision:%-8lu\n", FEC_IEEE_T_1COL);
	printf(" Tx >1 Collision:%-8lu", FEC_IEEE_T_MCOL);
	printf("     Tx Deferred:%-8lu", FEC_IEEE_T_DEF);
	printf("  Late Collision:%-8lu\n", FEC_IEEE_T_LCOL);
	printf("  Excessive Coll:%-8lu", FEC_IEEE_T_EXCOL);
	printf("   FIFO Underrun:%-8lu", FEC_IEEE_T_MACERR);
	printf("   Carrier Error:%-8lu\n", FEC_IEEE_T_CSERR);
	printf("    Tx SQE Error:%-8lu", FEC_IEEE_T_SQE);
	printf(" Tx Pause Frames:%-8lu", FEC_IEEE_T_FDXFC);
	printf("    Tx Octets OK:%-8lu\n", FEC_IEEE_T_OCTETS_OK);
}

NBUF *Xbuf;

/* getXmitBuffer():
 * Return a pointer to the buffer that is to be used for transmission of
 * the next packet.  Since the monitor's driver is EXTREMELY basic,
 * there will only be one packet ever being transmitted.  No need to queue
 * up transmit packets.
 */
uchar *
getXmitBuffer(void)
{
	Xbuf = nbuf_tx_allocate();
	if(Xbuf)
	{
		return(Xbuf->data);
	}
	return 0;
}

/* sendBuffer():
 * Send out the packet assumed to be built in the buffer returned by the
 * previous call to getXmitBuffer() above.
 */
int
sendBuffer(int length)
{
#if INCLUDE_ETHERVERBOSE
	if (EtherVerbose &  SHOW_OUTGOING)
		printPkt((struct ether_header *)Xbuf->data,length,ETHER_OUTGOING);

#endif

	WATCHDOG_MACRO;

	// Bump up the packet length to a minimum of 64 bytes.
	//if (length < 64)
	//	length = 64;
	 
	/* Add the code that will tickle the device into sending out the
	 * buffer that was previously returned by getXmitBuffer() above...
	 */
	Xbuf->length = length;

	/* Set Frame ready for transmit */
	Xbuf->status |= (FEC_TxBD_R | FEC_TxBD_TC);

	/* Mark the buffer as not in use */
	nbuf_tx_release(Xbuf);

	//printf("Xbuf->length: 0x%08lx\n", Xbuf->length);
	//printf("Xbuf->status: 0x%08lx\n", Xbuf->status);
	//printf("Xbuf->data: 0x%08lx\n", Xbuf->data);
	
	/* Indicate that there has been a transmit buffer produced */
	FEC_TDAR = FEC_TDAR_X_DES_ACTIVE;
	
	//udelay(100);
	//printf("Xbuf->length1: 0x%08lx\n", Xbuf->length);
	//printf("Xbuf->status1: 0x%08lx\n", Xbuf->status);
	//printf("Xbuf->data1: 0x%08lx\n", Xbuf->data);

	/* The hardware will mark the buffer as empty once it's transmitted */

	EtherXFRAMECnt++;
	return(length);
}

/* DisableEtherdev():
 * Fine as it is...
 */
void
DisableEtherdev(void)
{
	enreset();
}

/* extGetIpAdd():
 * If there was some external mechanism (other than just using the
 * IPADD shell variable established in the monrc file) for retrieval of
 * the board's IP address, then do it here...
 */
char *
extGetIpAdd(void)
{
	return((char *)0);
}

/* extGetEtherAdd():
 * If there was some external mechanism (other than just using the
 * ETHERADD shell variable established in the monrc file) for retrieval of
 * the board's MAC address, then do it here...
 */
char *
extGetEtherAdd(void)
{
	return((char *)0);
}

/*
 * polletherdev():
 * Called continuously by the monitor (ethernet.c) to determine if there
 * is any incoming ethernet packets.
 *
 * NOTES:
 * 1. This function must be reentrant, because there are a few cases in
 *    processPACKET() where pollethernet() may be called.
 * 2. It should only process one packet per call.  This is important
 *    because if allowed to stay here to flush all available packets,
 *    it may starve the rest of the system (especially in cases of heavy
 *    network traffic).
 * 3. There are cases in the monitor's execution that may cause the
 *    polling polletherdev() to cease for several seconds.  Depending on
 *    network traffic, this may cause the input buffering mechanism on
 *    the ethernet device to overflow.  A robust polletherdev() function
 *    should support this gracefully (i.e. when the error is detected,
 *    attempt to pass all queued packets to processPACKET(), then do what
 *    is necessary to clear the error).
 */
int
polletherdev(void)
{
	int pktcnt = 0;
	NBUF *pNbuf;

	WATCHDOG_MACRO;

	if (nbuf_rx_next_ready()) {
		/* Get the next buffer received by the FEC */
		pNbuf = nbuf_rx_allocate();
		if (pNbuf == 0) {
			printf("rx_allocate1 failed\n");
			return(pktcnt);
		}
		if (!(pNbuf->status & FEC_RxBD_L))
		{
			do
			{	/* Free the current buffer */
				nbuf_rx_release(pNbuf);

				/* Get the next buffer of this frame */
				pNbuf = nbuf_rx_allocate();
				if (pNbuf == 0) {
					printf("rx_allocate2 failed\n");
					return(pktcnt);
				}
	
				/* Keep going until we find the last one */
			} while (!(pNbuf->status & FEC_RxBD_L));

			/* Free the last buffer of this frame */
			nbuf_rx_release(pNbuf);
			return(pktcnt);
		}
	
		/* Call the appriopriate handler */
		processPACKET((struct ether_header *) pNbuf->data, pNbuf->length);
	
		nbuf_rx_release(pNbuf);
		pktcnt++;

		FEC_RDAR = FEC_RDAR_R_DES_ACTIVE;
	}
	return(pktcnt);
}

#endif
