/* reset.S:
 *	MPR2 (SH7720 based) board initial instructions.
 *	
 *	The reset vector point of the SH7720 is assumed to contain the resetPC (coldstart)
 *	followed by a stack pointer (MonStackEnd).
 *	The SH7720 reset vector is a pointer, not an instruction. Because
 *	of this, the first byte of binary in the normal image of the monitor can
 *	not be jumped to. When the monitor image is built to execute as a result
 *	of the boot-block CRC failure, then the first instruction in the image
 *	must be instruction. 
 *
 *	General notice:
 *	This code is part of a boot-monitor package developed as a generic base
 *	platform for embedded system designs.  As such, it is likely to be
 *	distributed to various projects beyond the control of the original
 *	author.  Please notify the author of any enhancements made or bugs found
 *	so that all may benefit from the changes.  In addition, notification back
 *	to the author will allow the new user to pick up changes that may have
 *	been made by other users after this version of the code was distributed.
 *
 *	Author:	Torsten Tetz, Bosch Security Systems
 *	        Mark Jonas, Bosch Security Systems, mark.jonas@de.bosch.com
 *	based on evalsh2/reset.S file by Ed Sutter
 */
#include "warmstart.h"
#include "config.h"

	.file "reset.s"

	.import	_start
	.globl coldstart, _coldstart, _warmstart
	.globl _moncomptr
	.globl _etheraddr

	.text

	.align	2

coldstart:
_coldstart:
	/* Set CPU to use RegisterBank 0 */
	mov.l 	SR_RB,r0
	not	r0,r0
	stc	sr,r1
	and	r0,r1
	ldc	r1,sr

	/* Continue behind moncomptr symbol */
	bra _coldstart_continue
	nop
	nop
	nop
	nop

	/* Maintain the 0x20 offset for _moncom: */
	.long 0x00000000, 0x00000000, 0x00000000

	/* Pointer to the moncom function used by application: */
_moncomptr:
	.long	_moncom

	/* Provide space to allow a programmer to place an ascii
	 * string in each of these locations as an optional point
	 * of storage for MAC and/or IP address...
	 */
_etheraddr:
	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	.byte 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff

_coldstart_continue:
	/* Load r4 with INITIALIZE */
	mov.l INITVAL,r4

_warmstart:
	/*
	 * Set chip frequency I:B:P. 
	 */

	/* Disable internal watchdog timer (WDT). */
	mov.l	WTCSR,r0		/* *WTCSR = WTCSR_TME_VAL; */
	mov.w	WTCSR_TME_VAL,r1
	mov.w	r1,@r0

	/* Set clock divider for watchdog timer (WDT). */
	mov.w	WTCSR_CKS_VAL,r1	/* *WTCSR = WTCSR_CKS_VAL; */
	mov.w	r1,@r0

	/* Set initial value for "Watchdog Timer Counter" */
	mov.l	WTCNT,r0		/* *WTCNT = WTCNT_VAL; */
	mov.w	WTCNT_VAL,r1
	mov.w	r1,@r0

	/* Set new frequency values. */
	mov.l	FRQCR,r0		/* *FRQCR = FRQCR_VAL; */
	mov.w	FRQCR_VAL,r1
	mov.w	r1,@r0

	/*
	 * Chip select settings 
	 */

	/* Set bus control register for CS3 */
	mov.l	CS3BCR,r0
	mov.l	CS3BCR_VAL,r1
	mov.l	r1,@r0

	/* Set wait control register for CS3 */
	mov.l	CS3WCR,r0
	mov.l	CS3WCR_VAL,r1
	mov.l	r1,@r0
	
	/*
	 * SDRAM controller timing settings 
	 */

	/* Set SDRAM control register */
	mov.l	SDCR,r1
	mov.l	SDCR_VAL,r2
	mov.l	r2,@r1

	/* Set refresh timer control/status register */
	mov.l	RTCSR,r1
	mov.l	RTCSR_VAL,r2
	mov.l	r2,@r1
	
	/* Set refresh timer counter */
	mov.l	RTCNT,r1
	mov.l	RTCNT_VAL,r2
	mov.l	r2,@r1

	/* Set refresh timer constant register */
	mov.l	RTCOR,r1
	mov.l	RTCOR_VAL,r2
	mov.l	r2,@r1

	/* Set SDRAM refresh control to auto-refresh */
	mov.l	SDCR_VAL,r1		/* *SDCR = SDCR_VAL | RFSH_VAL */
	mov.l	RFSH_VAL,r2
	or	r1,r2
	mov.l	SDCR,r1
	mov.l	r2,@r1

	/* Wait for 8 refresh periods */
	mov	#8,r3
refresh_count_loop:

	/* Clear refresh flag */
	mov.l	RTCSR,r1		/* *RTCSR = RTCSR_CWP | (*RTCSR & ~RTCSR_CMF) */
	mov.l	RTCSR_CMF,r2
	not	r2,r2
	mov.l	@r1,r0
	and	r0,r2
	mov.l	RTCSR_CWP,r0
	or	r0,r2
	mov.l	r2,@r1

	/* Wait till compare match flag CMF indicates refresh */
refresh_poll_loop:
	mov.l	RTCSR,r1		/* while (!(*RTCSR & RTCSR_CMF)) */
	mov.l	RTCSR_CMF,r2
	mov.l	@r1,r1
	and	r1,r2
	tst	r2,r2
	bt	refresh_poll_loop

	dt	r3			/* while (--R3 != 0) */
	bf	refresh_count_loop	

	/* Set SDRAM chip mode for area 3 */
	mov.l	SDMR3,r1		/* *((volatile unsigned short *) (SDMR3 | SDMR3_VAL)) = 0x0000 */
	mov.l	SDMR3_VAL,r2
	or	r1,r2
	mov.w	r1,@r2

	/* Flush cache.
	 * When uMon was restarted by jumping to "coldstart" above, the
	 * cache still contains old values. 
	 */
	mov.l	CCR1,r1
	mov.l	CCR1_FVAL,r2
	mov.l	r2,@r1

	/* Flush TLB */
	mov.l	MMUCR,r1
	mov.l	MMUCR_FVAL,r2
	mov.l	r2,@r1

	/* Load stack pointer. */
	mov.l	StackTag,r15
	mov.l	StackSize,r1
	add	r1,r15

	/* Jump into C code... */
	mov.l	StartTag,r8
	jmp		@r8
	nop

	.align 4 
StartTag:
	.long	_start

StackTag:
	.long 	_MonStack

StackSize:
	.long	MONSTACKSIZE

INITVAL:
	.long	INITIALIZE

CCR1:
	.long	0xFFFFFFEC
CS3BCR:
	.long	0xA4FD000C
CS3WCR:
	.long	0xA4FD002C
FRQCR:
	.long	0xA415FF80
MMUCR:
	.long	0xFFFFFFE0
RTCNT:
	.long	0xA4FD004C
RTCOR:
	.long	0xA4FD0050
RTCSR:
	.long	0xA4FD0048
SDCR:
	.long	0xA4FD0044
SDMR3:
	.long	0xA4FD5000
SR_RB:
	.long	0x20000000
WTCNT:				/* Use a word access to write to WTCNT, with H'5A in the upper byte. */
	.long	0xA415FF84
WTCSR:				/* Use a word access to write to WTCSR, with H'A5 in the upper byte. */
	.long	0xA415ff86


CCR1_FVAL:
	.long	0x00000008	/* CF bit: flus all cache entries */
CS3BCR_VAL:
	.long	0x10004400	/* minimum idle cycles, SDRAM, 16 bit */
MMUCR_FVAL:
	.long	0x00000004	/* TF bit: TLB flush */
RFSH_VAL:
	.long	0x00000800	/* refresh */
RTCSR_CMF:
	.long	0x00000080	/* RTCSR CMF bit */
RTCSR_CWP:
	.long	0xA55A0000	/* upper 16 bits of RTCSR must be 0xA55A to cancel write protect */
RTCSR_VAL:
	.long	0xA55A0008	/* clock select 1/4 */
SDCR_VAL:
	.long	0x00000012	/* no refresh, 13 rows, 10 cols, NO bank active mode */
SDMR3_VAL:
	.long	0x440		/* mode register CL2, burst read and SINGLE WRITE */

WTCSR_TME_VAL:
	.word	0xA57F		/* watchdog timer disable */
WTCSR_CKS_VAL:
	.word	0xA507		/* use peripheral clock/4096 (maximum) */
WTCNT_VAL:
	.word	0x5A00		/* initialize register to H'00 */

	.align 4

#if (QUARTZ==11000000)
CS3WCR_VAL:
	.long	0x00000090	/* tRP=1ck, tRCD=1ck, CL=2, tRWL=2ck, tRC=3ck */
RTCNT_VAL:
	.long	0xA55A0056	/* count 86 */
RTCOR_VAL:
	.long	0xA55A0056	/* count 86 */
FRQCR_VAL
	.word	0x1204		/* set I:B:P to 12:4:2. */
#elif (QUARTZ==12000000)
CS3WCR_VAL:
	.long	0x00000091	/* tRP=1ck, tRCD=1ck, CL=2, tRWL=2ck, tRC=4ck */
RTCNT_VAL:
	.long	0xA55A005D	/* count 93 */
RTCOR_VAL:
	.long	0xA55A005D	/* count 93 */
FRQCR_VAL:
	.word	0x1103		/* set I:B:P to 8:4:2. */
#elif (QUARTZ==16500000)
CS3WCR_VAL:
	.long	0x00002492	/* tRP=2ck, tRCD=2ck, CL=2, tRWL=2ck, tRC=6ck */
RTCNT_VAL:
	.long	0xA55A0082	/* count 130 */
RTCOR_VAL:
	.long	0xA55A0082	/* count 130 */
FRQCR_VAL:
	.word	0x1103		/* set I:B:P to 8:4:2. */
#else
   #error QUARTZ has an unsupported value.
#endif
	.align 4
