/* uart.c:
 *	Code for using the serial port monitor of the MPR2 board.
 *
 *	General notice:
 *	This code is part of a boot-monitor package developed as a generic base
 *	platform for embedded system designs.  As such, it is likely to be
 *	distributed to various projects beyond the control of the original
 *	author.  Please notify the author of any enhancements made or bugs found
 *	so that all may benefit from the changes.  In addition, notification back
 *	to the author will allow the new user to pick up changes that may have
 *	been made by other users after this version of the code was distributed.
 *
 *	Author:	Torsten Tetz, Bosch Security Systems
 *	        Mark Jonas, Bosch Security Systems, mark.jonas@de.bosch.com
 */
#include "config.h"
#include "timer.h"
#include "cpuio.h"
#include <limits.h>
#include "uart.h"

/* initUART():
 * This is the top-level UART initialization for the monitor.
 * This function is called at startup, application exit, or by the 
 * set -b BAUD command.
 * For startup and set, the baud rate will be specified and used; for
 * application exit, it will be 0 to indicate that whatever baud rate was
 * used prior by set or startup should be used again.
 */
void
initUART(int baud)
{
	/* Set transmit data transfer interrupt request is issued to CPU.
	 * Set receive data transfer interrupt request is issued to CPU.
	 * Set the transmit-data-stop-interrupt to disabled.
	 * Set the receive-error interrupt to disabled.
	 * Set the break-receive interrupt to disabled.
	 * Set the receive-data-ready interrupt to disabled.
	 * Set transmit-FIFO-data-empty interrupt request to disabled.
	 * Set receive-FIFO-data-full interrupt request to disabled.
	 * Disable the SCIF serial transmitter.
	 * Disable the SCIF serial receiver.
	 */
	*SCSCR_0 = (*SCSCR_0 & ~(SCSCR_TDRQE | SCSCR_RDRQE | SCSCR_TSIE | SCSCR_ERIE | SCSCR_BRIE | SCSCR_DRIE | SCSCR_TIE | SCSCR_RIE | SCSCR_TE | SCSCR_RE));

	/* Reset through TCRST, TFRST, RFRST */
	*SCFCR_0 = (*SCFCR_0 | SCFCR_TCRST | SCFCR_TFRST | SCFCR_RFRST);	

	/* Select the internal clock. */
	*SCSCR_0 = (*SCSCR_0 & ~(SCSCR_CKE0 | SCSCR_CKE1));			

	/* Set the baud rate. */
	ConsoleBaudSet(baud);

	/* TTRG=b'11 */
	*SCFCR_0 = (0x0000 | SCFCR_TTRG_0 | SCFCR_TTRG_1);

	/* Enable the SCIF serial transmitter and receiver. */
	*SCSCR_0 = (*SCSCR_0 | SCSCR_TE | SCSCR_RE);	
}

/* ConsoleBaudSet():
 * The common version of env.c calls ConsoleBaudSet()
 * in an attempt to adjust the console baudrate.
 *
 * return  0  when Baud rate was set succesfully.
 * return -1  when Baud rate could not be set.
 */
int
ConsoleBaudSet(int baud)
{
	const int numDiv = 8;
	const int numClocks = 4;
	const int div[8] = {16, 5, 7, 11, 13, 17, 19, 27};
	int brr;
	const int fPeri = PERICLCKFREQ;
	int di_min = -1;
	int n_min = -1;
	int brr_min = -1;
	int diff_min = INT_MAX;
	int di, n, n2;
	int denominator;
	int actDenominator;
	int baudrate;
	int diff;
	struct elapsed_tmr tmr;

	/* Do not continue if peripheral clock is not available. */
	if(fPeri == 0) {
		return(-1);
	}

	/*
	 * Find the setting that is the best match for the baud rate.
	 */

	/* Go over all sampling rates. */
	for(di=0; di < numDiv; di++) {

		/* Go over all clock sources. */
		for(n=0, n2=1; n < numClocks; n++, n2*=4) {

			/* The formula to calculate the Baud rate register setting is taken
			 * from the SH7720 user manual and was modified to work with integers:
			 *    brr = (fPeri / (SamplingRate * 2^(2*n) * BaudRate)) - 1
			 * Note that the result of the division needs to be rounded properly.
			 */
			denominator = div[di] * n2 * baud;

			/* Note the trick for fixed point rounding. */
			brr = ((fPeri + (denominator / 2)) / denominator) - 1;

			if((brr >= 0) && (brr <= 255)) {
				/* Calculate actual Baud rate this setting would yield. */
				actDenominator = div[di] * n2 * (brr + 1);

				/* Note the trick for fixed point rounding. */
				baudrate = ((fPeri + (actDenominator / 2)) / actDenominator);

				/* Calculate absolute difference to target Baud rate. */
				diff = baudrate - baud;
				if(diff < 0) {
					diff = 0 - diff;
				}

				/* If this setting's difference is smaller than anything before, keep it. */
				if(diff < diff_min) {
					di_min = di;
					n_min = n;
					brr_min = brr;
					diff_min = diff;
				}
			}
		}
	}

	/* Test the result. */
	if((di_min == -1) || (n_min == -1) || (brr_min == -1)) {
		return(-1);
	}

	/* Set samping rate. */
	*SCSMR_0 = ((*SCSMR_0 & ~SCSMR_SRC) | (di_min<<SCSMR_SRC_SHIFT));

	/* Select clock source. */
	*SCSMR_0 = ((*SCSMR_0 & ~SCSMR_CKS) | n_min);

	/* Set Baud rate register. */
	* SCBRR_0 = brr_min; 

	/* Sleep at least 1 bit time, so 1 ms is sufficient for 100 Baud and higher. */
	startElapsedTimer(&tmr,1);
	while(!msecElapsed(&tmr)) { }

	return(0);
}

/* target_console_empty():
 * Target-specific portion of flush_console() in chario.c.
 * This function returns 1 if there are no characters waiting to
 * be put out on the UART; else return 0 indicating that the UART
 * is still busy outputting characters from its FIFO.
 */
int
target_console_empty(void)
{
	/* Wait for empty tx buffer. */
	if(*SCSSR_0 & SCSSR_TDFE) {
		return(1);
	}
	return(0);
}

/* target_putchar():
 * Target-specific portion of character output to console used by
 * putchar() in chario.c.
 */
int
target_putchar(char c)
{
	/* Wait for empty rx buffer. */
	while((*SCFDR_0 & 0x7F00) != 0) { }

	/* Write char to rx buffer. */
	*SCFTDR_0 = (char) c;
	return(c);
}

/* target_getchar():
 * Target-specific portion of character input used by getchar()
 * in chario.c.  This function should assume that when it is called,
 * the caller has already polled on gotachar() waiting for a character
 * presence; hence, this function only needs to return the character.
 */
int
target_getchar(void)
{
	while(0 == (*SCFDR_0 & 0x007F)) { }
	return(*SCFRDR_0);
}

/* target_gotachar():
 * Return 0 if no char is avaialable at UART rcv FIFO; else 1.
 * Do NOT pull character out of FIFO, just return status.
 */
int
target_gotachar(void)
{
	if(*SCFDR_0 & 0x007F) {
		return(1);
	}
	else {
		return(0);
	}
}

